#ifndef ASM_E2K_LATENCY_INL_H
#define ASM_E2K_LATENCY_INL_H

#include "macro.h"
#include "latency.h"
#include "opcodes-e2k.h"

namespace asm_e2k {

const int8_t wr_reg_latency[][IN_REG_LAT_LAST] = {
#if __iset__ == 6
  // IOT_I  IOT_F  IOT_S IOT_W  IOT_R
  {     1,     1,     1,    1,  2 },   // OOT_0I,
  {     2,     2,     1,    3,  4 },   // OOT_1I,
  {     2,     2,     1,    3,  4 },   // OOT_1A,
  {     5,     5,     5,    5,  5 },   // OOT_2I,
  {     5,     5,     4,    4,  5 },   // OOT_4I,
  {     4,     1,     1,    3,  4 },   // OOT_1M,
  {     4,     2,     2,    3,  4 },   // OOT_2M,
  {     6,     4,     4,    6,  6 },   // OOT_4F,
  {     7,     5,     5,    7,  7 },   // OOT_5F,
  {     8,     6,     6,    8,  8 },   // OOT_6F,
  {     9,     9,     8,    8,  9 },   // OOT_8R,
  {    10,     8,     8,   10, 10 },   // OOT_8F,
  {    11,     9,     9,   11, 11 },   // OOT_9F,
  {    12,    10,    10,   12, 12 },   // OOT_10F,
  {    12,    10,    10,   12, 12 },   // OOT_10_14F,
  {    13,    11,    11,   13, 13 },   // OOT_11F,
  {    13,    11,    11,   13, 13 },   // OOT_11_14F,
  {    13,    11,    11,   13, 13 },   // OOT_11_16F,
  {    13,    11,    11,   13, 13 },   // OOT_11_18F,
  {    14,    12,    12,   14, 14 },   // OOT_12F,
  {    15,    13,    13,   15, 15 },   // OOT_13F,
  {    16,    14,    14,   16, 16 },   // OOT_14F,
  {    18,    16,    16,   18, 18 },   // OOT_16F,
  {    16,    14,    14,   16, 16 },   // OOT_14_16F,
  {    20,    18,    18,   20, 20 },   // OOT_18F,
  {    33,    31,    31,   33, 33 },   // OOT_31F,
  {    20,    20,    19,   19, 20 },   // OOT_19_R,
#else
  // IOT_I  IOT_F  IOT_S IOT_W  IOT_R
  {     1,     1,     1,    1,  2 },   // OOT_0I,
  {     2,     2,     1,    3,  4 },   // OOT_1I,
  {     2,     2,     1,    3,  4 },   // OOT_1A,
  {     3,     3,     4,    4,  4 },   // OOT_2I,
  {     5,     5,     4,    4,  5 },   // OOT_4I,
  {     4,     1,     1,    3,  4 },   // OOT_1M,
  {     4,     2,     2,    3,  4 },   // OOT_2M,
  {     6,     4,     4,    6,  6 },   // OOT_4F,
  {     7,     5,     5,    7,  7 },   // OOT_5F,
  {     8,     6,     6,    8,  8 },   // OOT_6F,
  {     9,     9,     8,    8,  9 },   // OOT_8R,
  {    10,     8,     8,   10, 10 },   // OOT_8F,
  {    11,     9,     9,   11, 11 },   // OOT_9F,
  {    12,    10,    10,   12, 12 },   // OOT_10F,
  {    12,    10,    10,   12, 12 },   // OOT_10_14F,
  {    13,    11,    11,   13, 13 },   // OOT_11F,
  {    13,    11,    11,   13, 13 },   // OOT_11_14F,
  {    13,    11,    11,   13, 13 },   // OOT_11_16F,
  {    13,    11,    11,   13, 13 },   // OOT_11_18F,
  {    14,    12,    12,   14, 14 },   // OOT_12F,
  {    15,    13,    13,   15, 15 },   // OOT_13F,
  {    16,    14,    14,   16, 16 },   // OOT_14F,
  {    18,    16,    16,   18, 18 },   // OOT_16F,
  {    16,    14,    14,   16, 16 },   // OOT_14_16F,
  {    20,    18,    18,   20, 20 },   // OOT_18F,
  {    33,    31,    31,   33, 33 },   // OOT_31F,
  {    20,    20,    19,   19, 20 },   // OOT_19_R,
#endif
};

const int8_t wr_latency_predicates[][IN_PR_LAT_LAST] = {
  // CT(COND)  RLP    ELP  CT(CALL)
  // PUTTSD    MERGE
  // INVTS
  {        3,     2,   1,        0 },   // OOT_INT,
  {        5,     4,   3,        0 },   // OOT_FLOAT,
};

#define LATENCY_CT_COLUMNS 3

const int8_t wr_latency_ct[][LATENCY_CT_COLUMNS] = {
  // CT(ctpr)  DONE(ctpr)  RR
  {        2,         -1,  1 },   // DISP, LDISP, SDISP, GETTSD,
  {        5,         -1,  4 },   // GETPL, MOVTd, GETTC,
  {        2,         -1,  6 },   // RETURN,
  {        1,         -1,  1 },   // CT_proc,
  {        5,          5,  4 },   // RW,
};

int8_t reg_latency(OutRegLantencyType out_type, InRegLantencyType in_type, bool bypass) {
    ASM_E2K_ASSERT_MSG(in_type < IN_REG_LAT_LAST, "invalid in_type");
    if (!bypass) {
      if (in_type == IN_REG_LAT_S) {
        in_type = IN_REG_LAT_W;
      } else {
        in_type = IN_REG_LAT_R;
      }
    }
    int8_t res = wr_reg_latency[out_type][in_type];
    ASM_E2K_ASSERT_MSG(res != -1, "bad latency");
    return res;
}

int8_t pred_latency(E2kOpCode src, E2kOpCode trg) {

    OutRegLantencyType out_lat = alu_info[src].out_latency;
    ASM_E2K_ASSERT_MSG(out_lat == OUT_REG_LAT_PRF || out_lat == OUT_REG_LAT_PRI, "wrong output type for predicates dep");

    int row = (out_lat == OUT_REG_LAT_PRI) ? 0 : 1;

    int col = 1;
    if (trg == OP_CT) {
        col = 0;
    } else if (trg == OP_RLP) {
        col = 1;
    } else if (trg == OP_ELP) {
        col = 2;
    } else if (trg == OP_CALL) {
        col = 3;
    } else {
        col = 1;
    }

    return wr_latency_predicates[row][col];
}

int8_t ct_latency(E2kOpCode src, E2kOpCode trg) {
    int col = 0;
    if (trg == OP_RRd || trg == OP_RRs) {
        col = 2;
    }
    int row = 0;
    switch(src) {
    case OP_DISP:
        row = 0; break;
    case OP_MOVTs:
    case OP_MOVTd:
    case OP_GETPL:
    case OP_GETTC:
        row = 1; break;
    case OP_RETURN:
        row = 2;
    case OP_RWs:
    case OP_RWd:
        row = 4;
    }
    int8_t lat = wr_latency_ct[row][col];
    ASM_E2K_ASSERT(lat != -1);
    return lat;
}

int8_t inst_latency(E2kOpCode src, E2kOpCode trg, bool bypass) {

    ASM_E2K_ASSERT(src < OP_PSEUDO_LAST);
    ASM_E2K_ASSERT(trg < OP_PSEUDO_LAST);

#ifdef ASM_SCHED_DEBUG
   printf("Latency Evaluation for  src=%s, trg=%s\n",
           alu_info[src].mnem, alu_info[trg].mnem);
#endif
    if ((src == OP_MOVTd || src == OP_MOVTs || src == OP_RETURN || src == OP_DISP) &&
        (trg == OP_CT || trg == OP_CALL || trg == OP_RRd || trg == OP_RRs)) { // MOVTd is a unique case.
        return ct_latency(src, trg);
    }

    OutRegLantencyType out_lat = alu_info[src].out_latency;
    InRegLantencyType  in_lat = alu_info[trg].in_latency;

    ASM_E2K_ASSERT(out_lat != OUT_REG_LAT_LAST || (src > OP_LAST && src < OP_PSEUDO_LAST));
    ASM_E2K_ASSERT(in_lat != IN_REG_LAT_LAST || (trg > OP_LAST && trg < OP_PSEUDO_LAST));

    if (out_lat == OUT_REG_LAT_PRF || out_lat == OUT_REG_LAT_PRI) {
        return pred_latency(src, trg);
    }

    ASM_E2K_ASSERT_MSG(in_lat != IN_REG_LAT_CT, "iot_ct for regs");

    return reg_latency(out_lat, in_lat, bypass);
}

LatencyMap::LatencyMap() : cur_lat(0), now(0) {
    std::fill((uint8_t*)data, (uint8_t*)data + sizeof(data), 0);
    defs.num_defs = 0;
}

inline uint8_t LatencyMap::_cluster(uint16_t exec_unit) {
    uint8_t cluster = ((exec_unit & EU_CLUSTER1) ? 1 : 0) | ((exec_unit & EU_CLUSTER2) ? 2 : 0);
    return cluster;
}

int8_t LatencyMap::current_latency() {
    return cur_lat;
}

void LatencyMap::set_current_latency(int val) {
    cur_lat = val;
}

void LatencyMap::increment_latency(int8_t nops) {
    cur_lat = 0;
    now += 1 + nops;
    if (nops) {
        // Update defs ticks for current bundle
        for (int i = 0; i < defs.num_defs; i++) {
            data[defs.defs[i]].ticks += nops;
        }
    }
    defs.num_defs = 0; // Reset def list
#ifdef ASM_SCHED_DEBUG
    printf("increment_latency, now=%d\n", now);
#endif
}

void LatencyMap::def(E2kOpCode op, uint16_t exec_unit, uint32_t reg_code) {
    ASM_E2K_ASSERT(reg_code < EREG_LAST);
    uint8_t cluster = ((exec_unit & EU_CLUSTER1) ? 1 : 0) | ((exec_unit & EU_CLUSTER2) ? 2 : 0);
    Latency& dst = data[reg_code];
    dst.init = true;
    dst.cluster = cluster;
    dst.op = op;
    dst.ticks = now;
    defs.defs[defs.num_defs++] = (uint16_t)reg_code;
    ASM_E2K_ASSERT(defs.num_defs <= deflist_t::MAX_DEFS_PER_BUNDLE);
}

inline void LatencyMap::_use(E2kOpCode op, uint8_t cluster, uint32_t reg_code) {
    Latency& src = data[reg_code];
    if(src.init) {
        bool bypass = src.cluster == cluster;
        uint32_t i_latency = inst_latency(src.op, op, bypass);
#ifdef ASM_SCHED_DEBUG
       printf("bypass %d, use lat %d, src=%s, trg=%s, src.ticks=%d, now=%d\n", bypass, i_latency, alu_info[src.op].mnem, alu_info[op].mnem, src.ticks, now);
#endif
        if (now - src.ticks > i_latency) {
            cur_lat = 0;
            return;
        }
        int8_t latency = std::max((int8_t)0, (int8_t)(i_latency - (now - src.ticks)));
        cur_lat = std::max(cur_lat, latency);
    }
}

inline int32_t LatencyMap::get_lat(E2kOpCode op, uint16_t exec_unit, uint32_t reg_code) {
    uint8_t cluster = ((exec_unit & EU_CLUSTER1) ? 1 : 0) | ((exec_unit & EU_CLUSTER2) ? 2 : 0);
    Latency& src = data[reg_code];
    if(src.init) {
        bool bypass = src.cluster == cluster;
        uint32_t i_latency = inst_latency(src.op, op, bypass);
#ifdef ASM_SCHED_DEBUG
        printf("get_lat:: bypass %d, use lat %d, src=%s, trg=%s, src.ticks=%d, now=%d\n", bypass, i_latency, alu_info[src.op].mnem, alu_info[op].mnem, src.ticks, now);
#endif
        if (now - src.ticks > i_latency) {
            return 0;
        }
        int8_t latency = std::max((int8_t)0, (int8_t)(i_latency - (now - src.ticks)));
        return latency;
    }
    return 0;
}

inline int32_t LatencyMap::get_bypass_lat(E2kOpCode op, bool bypass, uint32_t reg_code) {
    Latency& src = data[reg_code];
    if(src.init) {
        uint32_t i_latency = inst_latency(src.op, op, bypass);
#ifdef ASM_SCHED_DEBUG
        printf("get_bypass_lat:: bypass %d, use lat %d, src=%s, trg=%s, src.ticks=%d, now=%d\n", bypass, i_latency, alu_info[src.op].mnem, alu_info[op].mnem, src.ticks, now);
#endif
        if (now - src.ticks > i_latency) {
            return 0;
        }
        int8_t latency = std::max((int8_t)0, (int8_t)(i_latency - (now - src.ticks)));
        return latency;
    }
    return 0;
}

inline int32_t LatencyMap::get_distance(uint32_t reg_code) {
    Latency& src = data[reg_code];
    if(src.init) {
        return now - src.ticks;
    }
    return -1;
}

void LatencyMap::use(E2kOpCode op, uint16_t exec_unit, int64_t use1_idx) {
    uint8_t cluster = _cluster(exec_unit);
    _use(op, cluster, use1_idx);
}

void LatencyMap::use(E2kOpCode op, uint16_t exec_unit, int64_t use1_idx, int64_t use2_idx) {
    uint8_t cluster = _cluster(exec_unit);
    _use(op, cluster, use1_idx);
    _use(op, cluster, use2_idx);
}

void LatencyMap::use(E2kOpCode op, uint16_t exec_unit, int64_t use1_idx, int64_t use2_idx, int64_t use3_idx) {
    uint8_t cluster = _cluster(exec_unit);
    _use(op, cluster, use1_idx);
    _use(op, cluster, use2_idx);
    _use(op, cluster, use3_idx);
}

void LatencyMap::use(E2kOpCode op, uint16_t exec_unit, int64_t use1_idx, int64_t use2_idx, int64_t use3_idx, int64_t use4_idx) {
    uint8_t cluster = _cluster(exec_unit);
    _use(op, cluster, use1_idx);
    _use(op, cluster, use2_idx);
    _use(op, cluster, use3_idx);
    _use(op, cluster, use4_idx);
}

void LatencyMap::use(E2kOpCode op, uint16_t exec_unit, const Operand& use1) {
    uint8_t cluster = _cluster(exec_unit);
    if(use1.isReg()) _use(op, cluster, use1.value.reg);
}

void LatencyMap::use(E2kOpCode op, uint16_t exec_unit, const Operand& use1, const Operand& use2) {
    uint8_t cluster = _cluster(exec_unit);
    if(use1.isReg()) _use(op, cluster, use1.value.reg);
    if(use2.isReg()) _use(op, cluster, use2.value.reg);
}

void LatencyMap::use(E2kOpCode op, uint16_t exec_unit, const Operand& use1, const Operand& use2, const Operand& use3) {
    uint8_t cluster = _cluster(exec_unit);
    if(use1.isReg()) _use(op, cluster, use1.value.reg);
    if(use2.isReg()) _use(op, cluster, use2.value.reg);
    if(use3.isReg()) _use(op, cluster, use3.value.reg);
}

void LatencyMap::use(E2kOpCode op, uint16_t exec_unit, const Operand& use1, const Operand& use2, const Operand& use3, const Operand& use4) {
    uint8_t cluster = _cluster(exec_unit);
    if(use1.isReg()) _use(op, cluster, use1.value.reg);
    if(use2.isReg()) _use(op, cluster, use2.value.reg);
    if(use3.isReg()) _use(op, cluster, use3.value.reg);
    if(use4.isReg()) _use(op, cluster, use4.value.reg);
}

} // namespace asm_e2k

#endif // ASM_E2K_LATENCY_INL_H

