#ifdef __cplusplus
#ifndef ASM_E2K_LATENCY_H
#define ASM_E2K_LATENCY_H

#include "stdint.h"
#include "opcodes-e2k.h"
#include "operands.h"
#include "registers.h"

namespace asm_e2k {

enum InRegLantencyType : uint8_t {
    IN_REG_LAT_I,
    IN_REG_LAT_F,
    IN_REG_LAT_S,
    IN_REG_LAT_W,    // IN_REG_LAT_S without bypass.
    IN_REG_LAT_R,
    IN_REG_LAT_LAST, // last register file type.
    IN_REG_LAT_CT,   // ct insts.
    IN_REG_LAT_PROC_CT
};

enum OutRegLantencyType : uint8_t {
    OUT_REG_LAT_0I,
    OUT_REG_LAT_1I,
    OUT_REG_LAT_1A,
    OUT_REG_LAT_2I,
    OUT_REG_LAT_4I,
    OUT_REG_LAT_1M,
    OUT_REG_LAT_2M,
    OUT_REG_LAT_4F,
    OUT_REG_LAT_5F,
    OUT_REG_LAT_6F,
    OUT_REG_LAT_8R,
    OUT_REG_LAT_8F,
    OUT_REG_LAT_9F,
    OUT_REG_LAT_10F,
    OUT_REG_LAT_10_14F,
    OUT_REG_LAT_11F,
    OUT_REG_LAT_11_14F,
    OUT_REG_LAT_11_16F,
    OUT_REG_LAT_11_18F,
    OUT_REG_LAT_12F,
    OUT_REG_LAT_13F,
    OUT_REG_LAT_14F,
    OUT_REG_LAT_14_16F,
    OUT_REG_LAT_16F,
    OUT_REG_LAT_18F,
    OUT_REG_LAT_31F,
    OUT_REG_LAT_19_R,
    OUT_REG_LAT_CT,  // ct insts.
    OUT_REG_LAT_PRI, // predicate int.
    OUT_REG_LAT_PRF, // predicate float.
    OUT_REG_LAT_LAST
};

enum InPredLatencyType {
    IN_PR_LAT_CT_COND,
    IN_PR_LAT_RLP,
    IN_PR_LAT_ELP,
    IN_PR_LAT_CT_CALL,
    IN_PR_LAT_LAST
};

enum OutPredLatencyType {
    OUT_PR_LAT_INT,
    OUT_PR_LAT_FLOAT
};

extern const int8_t wr_reg_latency[][IN_REG_LAT_LAST];

class LatencyMap {
    typedef struct {
        static const int MAX_DEFS_PER_BUNDLE = 10; // TODO: estimate
        uint16_t defs[MAX_DEFS_PER_BUNDLE];
        uint8_t num_defs;
    } deflist_t;
    deflist_t defs;
public:
    typedef uint32_t ticks_t;
    struct Latency {
        bool init : 1;
        uint8_t cluster : 2; // 1 - Cluster 1 (ALU 0/1/2), 2 Cluster 2 (ALU 3,4,5), 0 - None/Other
        E2kOpCode op : 29;
        ticks_t ticks;
    } data[TotalRegCount];

    LatencyMap();

    void def(E2kOpCode op, uint16_t exec_unit, uint32_t reg_code);
    void use(E2kOpCode op, uint16_t exec_unit, int64_t use1_idx);
    void use(E2kOpCode op, uint16_t exec_unit, int64_t use1_idx, int64_t use2_idx);
    void use(E2kOpCode op, uint16_t exec_unit, int64_t use1_idx, int64_t use2_idx, int64_t use3_idx);
    void use(E2kOpCode op, uint16_t exec_unit, int64_t use1_idx, int64_t use2_idx,
         int64_t use3_idx, int64_t  use4_idx);
    int get_lat(E2kOpCode op, uint16_t exec_unit, uint32_t reg_code);
    int get_bypass_lat(E2kOpCode op, bool bypass, uint32_t reg_code);
    int32_t get_distance(uint32_t reg_code);
    void use(E2kOpCode op, uint16_t exec_unit, const Operand& use1);
    void use(E2kOpCode op, uint16_t exec_unit, const Operand& use1, const Operand& use2);
    void use(E2kOpCode op, uint16_t exec_unit, const Operand& use1, const Operand& use2, const Operand& use3);
    void use(E2kOpCode op, uint16_t exec_unit, const Operand& use1, const Operand& use2, const Operand& use3, const Operand& use4);
    inline int8_t current_latency();
    inline void set_current_latency(int val);
    inline void increment_latency(int8_t nops);
private:
    inline void _use(E2kOpCode op, uint8_t cluster, uint32_t reg_code);
    inline uint8_t _cluster(uint16_t exec_unit);
    ticks_t now;
    int8_t cur_lat;
};


} // namespace asm_e2k

#endif // ASM_E2K_LATENCY_H
#endif // __cplusplus
