#ifndef ASM_E2K_MACRO_H//
#define ASM_E2K_MACRO_H

#define BIT(n) (1U << (n))

#ifdef __cplusplus

#define NAMESPACE_ASM_E2K_BEGIN namespace asm_e2k {
#define NAMESPACE_ASM_E2K_END } // namespace asm_e2k

#ifdef PRODUCT
# ifndef NDEBUG
#  define NDEBUG
# endif
#endif

//#undef NDEBUG
#ifndef NDEBUG
# define ASM_E2K_DEBUG 1
#endif

#include <stdio.h>
#include <assert.h>

#ifdef ASM_E2K_DEBUG
# define ASM_E2K_DEBUG_ONLY(a) a
#else
# define ASM_E2K_DEBUG_ONLY(a)
#endif

#define ASM_E2K_ASSERT(a) assert(a)
#define ASM_E2K_ASSERT_MSG(a, msg) assert(a && msg)

#define ASM_E2K_UNREACHABLE() __builtin_unreachable()

__attribute__((noreturn)) __inline void ASM_E2K_NORETURN() { ASM_E2K_UNREACHABLE(); while(1); }

#define ASM_E2K_CRASH(msg) do { fprintf(stderr, "%s:%d: %s\n", __FILE__, __LINE__, msg); raise(SIGSEGV); ASM_E2K_NORETURN(); } while(0)

#define ASM_E2K_NOT_IMPLEMENTED() ASM_E2K_CRASH("Not implemented!")

#define ASM_E2K_UNUSED(a) do { a = a; } while(0)

#ifndef ASM_E2K_RELOCATION_NONE
# define ASM_E2K_RELOCATION_NONE 0
#endif

#ifndef ASM_E2K_DBL_FLAG
# define ASM_E2K_DBL_FLAG 1
#endif

#ifndef ASM_E2K_AUTO_NOPS
# define ASM_E2K_AUTO_NOPS 1
#endif

#ifndef ASM_E2K_HARDCODED_NOPS
# define ASM_E2K_HARDCODED_NOPS 1
#endif

#ifndef ASM_E2K_LEGACY_GUARANTEE_DELAY
# define ASM_E2K_LEGACY_GUARANTEE_DELAY 0
#endif

#if ASM_E2K_LEGACY_GUARANTEE_DELAY && ASM_E2K_AUTO_NOPS
# warning "Both ASM_E2K_AUTO_NOPS and ASM_E2K_LEGACY_GUARANTEE_DELAY are enabled"
#endif

#define ASM_E2K_CHECK_EXTERNAL_BUFFER_OVERFLOW 0

//#undef ASM_E2K_GENERATE_C_API

#else // __cplusplus

#define NAMESPACE_ASM_E2K_BEGIN
#define NAMESPACE_ASM_E2K_END
#define ASM_E2K_GENERATE_C_API 1

#ifndef bool
# define bool _Bool
#endif

#endif // __cplusplus

#endif // ASM_E2K_MACRO_H
