#ifndef ASM_E2K_OPCODES_E2K_H
#define ASM_E2K_OPCODES_E2K_H

#include "macro.h"
#include "stdint.h"

NAMESPACE_ASM_E2K_BEGIN

typedef enum ElbrusOpcode {

    // Некомбинированные операции с короткой кодировкой
    OPC_ADDd      = 0x11,    OPC_ADDs      = 0x10,    OPC_ANDd      = 0x01,    OPC_ANDNd     = 0x03,
    OPC_ANDNs     = 0x02,    OPC_ANDs      = 0x00,    OPC_BITREVd   = 0x27,    OPC_BITREVs   = 0x26,
    OPC_CCTOB     = 0x24,    OPC_CMPANDdb  = 0x23,    OPC_CMPANDsb  = 0x22,    OPC_CMPdb     = 0x21,
    OPC_CMPsb     = 0x20,    OPC_DRTOAP    = 0x62,    OPC_FADDd     = 0x31,    OPC_FADDs     = 0x30,
    OPC_FCMPdb    = 0x2f,    OPC_FCMPsb    = 0x2e,    OPC_FDTOD     = 0x3d,    OPC_FDTOS     = 0x3f,
    OPC_FMAXd     = 0x37,    OPC_FMAXs     = 0x36,    OPC_FMINd     = 0x35,    OPC_FMINs     = 0x34,
    OPC_FMULd     = 0x39,    OPC_FMULs     = 0x38,    OPC_FSTOD     = 0x3e,    OPC_FSTOS     = 0x3c,
    OPC_FSUBd     = 0x33,    OPC_FSUBs     = 0x32,    OPC_FXADDdd   = 0x41,    OPC_FXADDdx   = 0x43,
    OPC_FXADDss   = 0x40,    OPC_FXADDsx   = 0x42,    OPC_FXADDxd   = 0x45,    OPC_FXADDxs   = 0x44,
    OPC_FXADDxx   = 0x47,    OPC_FXCMPdb   = 0x29,    OPC_FXCMPsb   = 0x28,    OPC_FXCMPxb   = 0x2b,
    OPC_FXDIVdd   = 0x49,    OPC_FXDIVdx   = 0x4b,    OPC_FXDIVss   = 0x48,    OPC_FXDIVsx   = 0x4a,
    OPC_FXDIVTdd  = 0x61,    OPC_FXDIVTdx  = 0x63,    OPC_FXDIVTss  = 0x60,    OPC_FXDIVTsx  = 0x62,
    OPC_FXDIVxd   = 0x4d,    OPC_FXDIVxs   = 0x4c,    OPC_FXDIVxx   = 0x4f,    OPC_FXMULdd   = 0x51,
    OPC_FXMULdx   = 0x53,    OPC_FXMULss   = 0x50,    OPC_FXMULsx   = 0x52,    OPC_FXMULxd   = 0x55,
    OPC_FXMULxs   = 0x54,    OPC_FXMULxx   = 0x57,    OPC_FXRSUBdd  = 0x59,    OPC_FXRSUBdx  = 0x5b,
    OPC_FXRSUBss  = 0x58,    OPC_FXRSUBsx  = 0x5a,    OPC_FXSQRTIdx = 0x53,    OPC_FXSQRTIsx = 0x52,
    OPC_FXSQRTIxx = 0x57,    OPC_FXSQRTTdx = 0x5f,    OPC_FXSQRTTsx = 0x5e,    OPC_FXSQRTTxx = 0x5d,
    OPC_FXSQRTUdx = 0x5b,    OPC_FXSQRTUsx = 0x5a,    OPC_FXSQRTUxx = 0x59,    OPC_FXSUBdd   = 0x49,
    OPC_FXSUBdx   = 0x4b,    OPC_FXSUBss   = 0x48,    OPC_FXSUBsx   = 0x4a,    OPC_FXSUBxd   = 0x4d,
    OPC_FXSUBxs   = 0x4c,    OPC_FXSUBxx   = 0x4f,    OPC_GETFd     = 0x1f,    OPC_GETFs     = 0x1e,
    OPC_GETPL     = 0x63,    OPC_LDB       = 0x64,    OPC_LDCSB     = 0x68,    OPC_LDCSD     = 0x6b,
    OPC_LDCSH     = 0x69,    OPC_LDCSW     = 0x6a,    OPC_LDD       = 0x67,    OPC_LDDSB     = 0x6c,
    OPC_LDDSD     = 0x6f,    OPC_LDDSH     = 0x6d,    OPC_LDDSW     = 0x6e,    OPC_LDESB     = 0x70,
    OPC_LDESD     = 0x73,    OPC_LDESH     = 0x71,    OPC_LDESW     = 0x72,    OPC_LDFSB     = 0x74,
    OPC_LDFSD     = 0x77,    OPC_LDFSH     = 0x75,    OPC_LDFSW     = 0x76,    OPC_LDGSB     = 0x78,
    OPC_LDGSD     = 0x7b,    OPC_LDGSH     = 0x79,    OPC_LDGSW     = 0x7a,    OPC_LDH       = 0x65,
    OPC_LDSSB     = 0x7c,    OPC_LDSSD     = 0x7f,    OPC_LDSSH     = 0x7d,    OPC_LDSSW     = 0x7e,
    OPC_LDW       = 0x66,    OPC_LZCNTd    = 0x65,    OPC_LZCNTs    = 0x64,    OPC_MERGEd    = 0x0f,
    OPC_MERGEs    = 0x0e,    OPC_MOVFI     = 0x5c,    OPC_MOVIF     = 0x5e,    OPC_MOVTd     = 0x61,
    OPC_MOVTs     = 0x60,    OPC_MOVX      = 0x5f,    OPC_ORd       = 0x05,    OPC_ORNd      = 0x07,
    OPC_ORNs      = 0x06,    OPC_ORs       = 0x04,    OPC_POPCNTd   = 0x67,    OPC_POPCNTs   = 0x66,
    OPC_SARd      = 0x1d,    OPC_SARs      = 0x1c,    OPC_SCLd      = 0x15,    OPC_SCLs      = 0x14,
    OPC_SCRd      = 0x17,    OPC_SCRs      = 0x16,    OPC_SDIVd     = 0x43,    OPC_SDIVs     = 0x42,
    OPC_SDIVX     = 0x46,    OPC_SHLd      = 0x19,    OPC_SHLs      = 0x18,    OPC_SHRd      = 0x1b,
    OPC_SHRs      = 0x1a,    OPC_SMODX     = 0x47,    OPC_STB       = 0x24,    OPC_STCSB     = 0x28,
    OPC_STCSD     = 0x2b,    OPC_STCSH     = 0x29,    OPC_STCSW     = 0x2a,    OPC_STD       = 0x27,
    OPC_STDSB     = 0x2c,    OPC_STDSD     = 0x2f,    OPC_STDSH     = 0x2d,    OPC_STDSW     = 0x2e,
    OPC_STESB     = 0x30,    OPC_STESD     = 0x33,    OPC_STESH     = 0x31,    OPC_STESW     = 0x32,
    OPC_STFSB     = 0x34,    OPC_STFSD     = 0x37,    OPC_STFSH     = 0x35,    OPC_STFSW     = 0x36,
    OPC_STGSB     = 0x38,    OPC_STGSD     = 0x3b,    OPC_STGSH     = 0x39,    OPC_STGSW     = 0x3a,
    OPC_STH       = 0x25,    OPC_STSSB     = 0x3c,    OPC_STSSD     = 0x3f,    OPC_STSSH     = 0x3d,
    OPC_STSSW     = 0x3e,    OPC_STW       = 0x26,    OPC_SUBd      = 0x13,    OPC_SUBs      = 0x12,
    OPC_SXT       = 0x0c,    OPC_UDIVd     = 0x41,    OPC_UDIVs     = 0x40,    OPC_UDIVX     = 0x44,
    OPC_UMODX     = 0x45,    OPC_VFSI      = 0x63,    OPC_XORd      = 0x09,    OPC_XORNd     = 0x0b,
    OPC_XORNs     = 0x0a,    OPC_XORs      = 0x08,

    // Расширения кода операции
    OPX_EXT       = 0x01,    OPX_FLB       = 0x04,    OPX_FLH       = 0x05,    OPX_FLW       = 0x06,
    OPX_FLD       = 0x07,    OPX_ICMB0     = 0x08,    OPX_ICMB1     = 0x09,    OPX_ICMB2     = 0x0a,
    OPX_ICMB3     = 0x0b,    OPX_FCMB0     = 0x0c,    OPX_FCMB1     = 0x0d,    OPX_PFCMB0    = 0x0e,
    OPX_PFCMB1    = 0x0f,    OPX_FXCMB     = 0x14,

    // Некомбинированные операции с длинной кодировкой
    OPL_APTOAP    = 0x50,    OPL_APTOAPB   = 0x51,    OPL_CAST      = 0x55,    OPL_FCMPd     = 0x2d,
    OPL_FCMPdf    = 0x2f,    OPL_FCMPs     = 0x2c,    OPL_FCMPsf    = 0x2e,    OPL_FDIVd     = 0x49,
    OPL_FDIVs     = 0x48,    OPL_FRCPs     = 0x50,    OPL_FRSQRTs   = 0x54,    OPL_FSQRTId   = 0x4d,
    OPL_FSQRTs    = 0x4c,    OPL_FSQRTTd   = 0x51,    OPL_FXCMPdf   = 0x29,    OPL_FXCMPsf   = 0x28,
    OPL_FXCMPxf   = 0x2b,    OPL_GETSOD    = 0x5a,    OPL_GETSP     = 0x58,    OPL_GETTAGd   = 0x09,
    OPL_GETTAGs   = 0x08,    OPL_GETTC     = 0x24,    OPL_GETTD     = 0x56,    OPL_GETVA     = 0x52,
    OPL_INVTC     = 0x26,    OPL_LDAAB     = 0x5c,    OPL_LDAAD     = 0x5f,    OPL_LDAAH     = 0x5d,
    OPL_LDAAQ     = 0x7f,    OPL_LDAAW     = 0x5e,    OPL_LDAPB     = 0x68,    OPL_LDAPD     = 0x6b,
    OPL_LDAPH     = 0x69,    OPL_LDAPQ     = 0x7a,    OPL_LDAPW     = 0x6a,    OPL_LDCSQ     = 0x42,
    OPL_LDCUDB    = 0x60,    OPL_LDCUDD    = 0x63,    OPL_LDCUDH    = 0x61,    OPL_LDCUDQ    = 0x78,
    OPL_LDCUDW    = 0x62,    OPL_LDDSQ     = 0x43,    OPL_LDESQ     = 0x44,    OPL_LDFSQ     = 0x45,
    OPL_LDGDB     = 0x64,    OPL_LDGDD     = 0x67,    OPL_LDGDH     = 0x65,    OPL_LDGDQ     = 0x79,
    OPL_LDGDW     = 0x66,    OPL_LDGSQ     = 0x46,    OPL_LDODPB    = 0x74,    OPL_LDODPD    = 0x77,
    OPL_LDODPH    = 0x75,    OPL_LDODPQ    = 0x7d,    OPL_LDODPW    = 0x76,    OPL_LDODRB    = 0x6c,
    OPL_LDODRD    = 0x6f,    OPL_LDODRH    = 0x6d,    OPL_LDODRQ    = 0x7b,    OPL_LDODRW    = 0x6e,
    OPL_LDODWB    = 0x70,    OPL_LDODWD    = 0x73,    OPL_LDODWH    = 0x71,    OPL_LDODWQ    = 0x7c,
    OPL_LDODWW    = 0x72,    OPL_LDRD      = 0x5b,    OPL_LDSSQ     = 0x47,    OPL_MOVTq     = 0x57,
    OPL_MULd      = 0x21,    OPL_MULs      = 0x20,    OPL_ODTOAP    = 0x54,    OPL_PACKSSHB  = 0x08,
    OPL_PACKSSWH  = 0x0a,    OPL_PACKUSHB  = 0x09,    OPL_PADDB     = 0x08,    OPL_PADDD     = 0x0f,
    OPL_PADDH     = 0x09,    OPL_PADDSB    = 0x0a,    OPL_PADDSH    = 0x0b,    OPL_PADDUSB   = 0x0c,
    OPL_PADDUSH   = 0x0d,    OPL_PADDW     = 0x0e,    OPL_PANDd     = 0x48,    OPL_PANDNd    = 0x49,
    OPL_PAVGUSB   = 0x1e,    OPL_PAVGUSH   = 0x1f,    OPL_PCMPEQB   = 0x18,    OPL_PCMPEQH   = 0x19,
    OPL_PCMPEQW   = 0x1a,    OPL_PCMPGTB   = 0x1b,    OPL_PCMPGTH   = 0x1c,    OPL_PCMPGTW   = 0x1d,
    OPL_PEXTRH    = 0x1e,    OPL_PFADDd    = 0x31,    OPL_PFADDs    = 0x30,    OPL_PFCMPd    = 0x3b,
    OPL_PFCMPs    = 0x3a,    OPL_PFDIVd    = 0x4b,    OPL_PFDIVs    = 0x4a,    OPL_PFMAXd    = 0x37,
    OPL_PFMAXs    = 0x36,    OPL_PFMINd    = 0x35,    OPL_PFMINs    = 0x34,    OPL_PFMULd    = 0x39,
    OPL_PFMULs    = 0x38,    OPL_PFSQRTs   = 0x4e,    OPL_PFSQRTTd  = 0x53,    OPL_PFSUBd    = 0x33,
    OPL_PFSUBs    = 0x32,    OPL_PINSH     = 0x1f,    OPL_PMADDH    = 0x1a,    OPL_PMAXSH    = 0x03,
    OPL_PMAXUB    = 0x02,    OPL_PMINSH    = 0x01,    OPL_PMINUB    = 0x00,    OPL_PMOVMSKB  = 0x0b,
    OPL_PMOVMSKPD = 0x07,    OPL_PMOVMSKPS = 0x06,    OPL_PMULHH    = 0x18,    OPL_PMULHUH   = 0x1b,
    OPL_PMULLH    = 0x19,    OPL_PMULUBHH  = 0x1d,    OPL_PORd      = 0x4a,    OPL_PSADBW    = 0x1c,
    OPL_PSHUFH    = 0x17,    OPL_PSHUFW    = 0x16,    OPL_PSLLD     = 0x4e,    OPL_PSLLH     = 0x15,
    OPL_PSLLQH    = 0x0e,    OPL_PSLLQL    = 0x0f,    OPL_PSLLW     = 0x14,    OPL_PSRAH     = 0x13,
    OPL_PSRAW     = 0x12,    OPL_PSRLD     = 0x4c,    OPL_PSRLH     = 0x11,    OPL_PSRLQH    = 0x0c,
    OPL_PSRLQL    = 0x0d,    OPL_PSRLW     = 0x10,    OPL_PSUBB     = 0x10,    OPL_PSUBD     = 0x17,
    OPL_PSUBH     = 0x11,    OPL_PSUBSB    = 0x12,    OPL_PSUBSH    = 0x13,    OPL_PSUBUSB   = 0x14,
    OPL_PSUBUSH   = 0x15,    OPL_PSUBW     = 0x16,    OPL_PUNPCKHBH = 0x00,    OPL_PUNPCKHHW = 0x02,
    OPL_PUNPCKHWD = 0x04,    OPL_PUNPCKLBH = 0x01,    OPL_PUNPCKLHW = 0x03,    OPL_PUNPCKLWD = 0x05,
    OPL_PUTTAGd   = 0x0b,    OPL_PUTTAGs   = 0x0a,    OPL_PUTTC     = 0x25,    OPL_PXORd     = 0x4b,
    OPL_RRd       = 0x3f,    OPL_RRs       = 0x3e,    OPL_RWd       = 0x3d,    OPL_RWs       = 0x3c,
    OPL_SMULH     = 0x71,    OPL_SMULX     = 0x23,    OPL_UMULH     = 0x70,    OPL_UMULX     = 0x22,
    OPL_STAAB     = 0x1c,    OPL_STAAD     = 0x1f,    OPL_STAAH     = 0x1d,
    OPL_STAAQ     = 0x3f,    OPL_STAAW     = 0x1e,    OPL_STAPB     = 0x28,    OPL_STAPD     = 0x2b,
    OPL_STAPH     = 0x29,    OPL_STAPQ     = 0x3a,    OPL_STAPW     = 0x2a,    OPL_STCSQ     = 0x02,
    OPL_STDSQ     = 0x03,    OPL_STESQ     = 0x04,    OPL_STFSQ     = 0x05,    OPL_STGDB     = 0x24,
    OPL_STGDD     = 0x27,    OPL_STGDH     = 0x25,    OPL_STGDQ     = 0x39,    OPL_STGDW     = 0x26,
    OPL_STGSQ     = 0x06,    OPL_STODPB    = 0x34,    OPL_STODPD    = 0x37,    OPL_STODPH    = 0x35,
    OPL_STODPQ    = 0x3d,    OPL_STODPW    = 0x36,    OPL_STODRB    = 0x2c,    OPL_STODRD    = 0x2f,
    OPL_STODRH    = 0x2d,    OPL_STODRQ    = 0x3b,    OPL_STODRW    = 0x2e,    OPL_STODWB    = 0x30,
    OPL_STODWD    = 0x33,    OPL_STODWH    = 0x31,    OPL_STODWQ    = 0x3c,    OPL_STODWW    = 0x32,
    OPL_STRD      = 0x1b,    OPL_STSSQ     = 0x07,    OPL_TDTOMP    = 0x53,    OPL_FDTOIFd   = 0x6d,

    // Операции над целыми с результатом в виде флага
    OPF_ANDs      = 0x00,    OPF_ANDNs     = 0x02,    OPF_ORs       = 0x04,    OPF_ORNs      = 0x06,
    OPF_XORs      = 0x08,    OPF_XORNs     = 0x0a,    OPF_ADDs      = 0x10,    OPF_SUBs      = 0x12,
    OPF_SCLs      = 0x14,    OPF_SCRs      = 0x16,    OPF_SHLs      = 0x18,    OPF_SHRs      = 0x1a,
    OPF_SARs      = 0x1c,    OPF_UMULX     = 0x22,    OPF_SMULX     = 0x23,    OPF_INCs      = 0x30,
    OPF_DECs      = 0x32,

    // Операции слога CS0
    OC0_IBRANCH   = 0x0,     OC0_DONE      = 0x3,     OC0_DISP      = 0x0,     OC0_LDISP     = 0x1,
    OC0_SDISP     = 0x2,     OC0_RETURN    = 0x3,

    // Операции слога CS1
    OC1_SETR0     = 0x00,    OC1_SETR1     = 0x01,    OC1_SETEI     = 0x02,    OC1_WAIT      = 0x03,
    OC1_SETBR     = 0x04,    OC1_CALL      = 0x05,    OC1_MAS       = 0x06,    OC1_FLUSHR    = 0x07,

    // Операции целых сравнений с результатом в виде предиката
    OCI_OVW       = 0x00,    OCI_LTU       = 0x01,    OCI_EQ        = 0x02,    OCI_LEU       = 0x03,
    OCI_NEG       = 0x04,    OCI_ODD       = 0x05,    OCI_LTS       = 0x06,    OCI_LES       = 0x07,

    // Операции вещественных сравнений с результатом в виде предиката
    OCF_EQ        = 0x00,    OCF_LT        = 0x01,    OCF_LE        = 0x02,    OCF_UO        = 0x03,
    OCF_NE        = 0x04,    OCF_NLT       = 0x05,    OCF_NLE       = 0x06,    OCF_OD        = 0x07,

    // Операции сравнения вещественных с результатом в виде битовой маски
    OCFM_EQ       = 0xc0,    OCFM_LT       = 0xc1,    OCFM_LE       = 0xc2,    OCFM_UO       = 0xc3,
    OCFM_NE       = 0xc4,    OCFM_NL       = 0xc5,    OCFM_NLE      = 0xc6,    OCFM_OD       = 0xc7,

    // Операции сравнения вещественных с результатом в виде флагов
    OCFF_OD       = 0xc7,    OCFF_UO       = 0xc3,

    // Операции преобразования вещественных/целых
    OCV_FDTOFS    = 0xc6,    OCV_FDTOFX    = 0xc7,    OCV_FDTOID    = 0xc0,    OCV_FDTOIDTR  = 0xc2,
    OCV_FDTOIS    = 0xc0,    OCV_FDTOISTR  = 0xc2,    OCV_FSTOFD    = 0xc6,    OCV_FSTOFX    = 0xc7,
    OCV_FSTOID    = 0xc0,    OCV_FSTOIDTR  = 0xc2,    OCV_FSTOIS    = 0xc0,    OCV_FSTOISTR  = 0xc2,
    OCV_FXTOFD    = 0xc6,    OCV_FXTOFS    = 0xc7,    OCV_FXTOID    = 0xc1,    OCV_FXTOIDTR  = 0xc3,
    OCV_FXTOIS    = 0xc1,    OCV_FXTOISTR  = 0xc3,    OCV_IDTOFD    = 0xc4,    OCV_IDTOFS    = 0xc4,
    OCV_IDTOFX    = 0xc5,    OCV_ISTOFD    = 0xc4,    OCV_ISTOFS    = 0xc4,    OCV_ISTOFX    = 0xc5,
    OCV_PFDTOFS   = 0xce,    OCV_PFDTOIS   = 0xc8,    OCV_PFDTOISTR = 0xca,    OCV_PFSTOFD   = 0xce,
    OCV_PFSTOIS   = 0xc8,    OCV_PFSTOISTR = 0xca,    OCV_PISTOFS   = 0xcc,    OCV_FDTOFXTOFD= 0xcf,
    OCV_FSTOFXTOFS= 0xcf,

    // Операции преобразования адресов
    OPA_NONE      = 0xc0,    OPA_CUD       = 0xf0,    OPA_GD        = 0xf2,    OPA_USD       = 0xec,
    OPA_STAT      = 0xee,

    // Операции Move для буфера предподкачки массива (AAS.opc)
    OAA_MOVAB     = 0x01,    OAA_MOVAH     = 0x02,    OAA_MOVAW     = 0x03,    OAA_MOVAD     = 0x04,
    OAA_MOVAQ     = 0x05,

    // Коды операций поля ALS.opce операций MOVT
    OPM_MV        = 0xc0,    OPM_MVC       = 0xc1,    OPM_MVR       = 0xc2,    OPM_MVRC      = 0xc3,

    // Коды операций поля ALS.opce операций INVTC
    OIV_TCA       = 0xc0,    OIV_TCE       = 0xc1,

    // Коды операций для значений по умолчанию полей ALS.opce и ALES.opce
    OPC_NONE      = 0xc0,

    // Коды операций поля ALS.opce операций MOVX
    OMX_MV        = 0xc0,    OMX_MVXA      = 0xc1,    OMX_MVXC      = 0xc2,

    // Коды операций SXT (SRC1)
    OSXT_BS       = 0x00,    OSXT_HS       = 0x01,    OSXT_WS       = 0x02,
    OSXT_BZ       = 0x04,    OSXT_HZ       = 0x05,    OSXT_WZ       = 0x06,

    // Коды операций полуслога CD
    OCD_RLP       = 0x00,    OCD_MRGC      = 0x02,

    // Коды операций слога PLS
    OPL_CLPAND    = 0x00,    OPL_CLPLAND   = 0x01,    OPL_MLP       = 0x03,

    OPC_LAST
} ElbrusOpcode;



enum ALUNum {
  ALU_0 = 0,
  ALU_1,
  ALU_2,
  ALU_3,
  ALU_4,
  ALU_5,

  ALU_TOTAL
};

#define EU_AL0 0x0001
#define EU_AL1 0x0002
#define EU_AL2 0x0004
#define EU_AL3 0x0008
#define EU_AL4 0x0010
#define EU_AL5 0x0020
#define EU_CS0 0x0040
#define EU_CS1 0x0080
#define EU_SS0 0x0100
#define EU_SS1 0x0200
#define EU_SS2 0x0400
#define EU_NOP 0x0800
#define EU_AA0 0x1000
#define EU_AA1 0x2000
#define EU_AA2 0x4000
#define EU_AA3 0x8000

typedef enum ExecUnitMask {
    EU_AL_012345 = EU_AL0 | EU_AL1 | EU_AL2 | EU_AL3 | EU_AL4 | EU_AL5,
    EU_AL_03     = EU_AL0 | EU_AL3,
    EU_AL_14     = EU_AL1 | EU_AL4,
    EU_AL_0134   = EU_AL0 | EU_AL1 | EU_AL3 | EU_AL4,
    EU_AL_0235   = EU_AL0 | EU_AL2 | EU_AL3 | EU_AL5,
    EU_AL_25     = EU_AL2 | EU_AL5,
    EU_AL_5      = EU_AL5,
    EU_AL_2      = EU_AL2,
    EU_AL_1      = EU_AL1,
    EU_AL_0      = EU_AL0,
    EU_CS_0      = EU_CS0,
    EU_CS_1      = EU_CS1,
    EU_SS_CTOP   = EU_SS0,
    EU_SS_BAP    = EU_SS1,
    EU_SS_ABN    = EU_SS2,
    EU_HS_NOP    = EU_NOP,
    EU_AA_0      = EU_AA0,
    EU_AA_1      = EU_AA1,
    EU_AA_2      = EU_AA2,
    EU_AA_3      = EU_AA3,
    EU_AA_0123   = EU_AA0 | EU_AA1 | EU_AA2 | EU_AA3,
    EU_UNLOCK_ST = EU_AL2 | EU_AL5 | EU_CS0,
    EU_ALL       = EU_AL_012345 | EU_CS0 | EU_CS1 | EU_SS0 | EU_SS1 | EU_SS2 | EU_NOP | EU_AA_0123,
    EU_CLUSTER1  = EU_AL0 | EU_AL1 | EU_AL2,
    EU_CLUSTER2  = EU_AL3 | EU_AL4 | EU_AL5,
    EU_LAST      = 0xFFFF
} ExecUnitMask;

typedef enum E2kOpCode {
    OP_ADDd = 0x0,//0x0
    OP_ADDs,      //0x1
    OP_ANDd,      //0x2
    OP_ANDNd,      //0x3
    OP_ANDNs,      //0x4
    OP_ANDs,      //0x5
    OP_BITREVd,      //0x6
    OP_BITREVs,      //0x7
    OP_CCTOB,      //0x8
    OP_CMPANDdb,      //0x9
    OP_CMPANDsb,      //0xa
    OP_CMPdb,      //0xb
    OP_CMPsb,      //0xc
    OP_DRTOAP,      //0xd
    OP_FADDd,      //0xe
    OP_FADDs,      //0xf
    OP_FCMPdb,      //0x10
    OP_FCMPsb,      //0x11
    OP_FDTOD,      //0x12
    OP_FDTOS,      //0x13
    OP_FMAXd,      //0x14
    OP_FMAXs,      //0x15
    OP_FMINd,      //0x16
    OP_FMINs,      //0x17
    OP_FMULd,      //0x18
    OP_FMULs,      //0x19
    OP_FSTOD,      //0x1a
    OP_FSTOS,      //0x1b
    OP_FSUBd,      //0x1c
    OP_FSUBs,      //0x1d
    OP_FXADDdd,      //0x1e
    OP_FXADDdx,      //0x1f
    OP_FXADDss,      //0x20
    OP_FXADDsx,      //0x21
    OP_FXADDxd,      //0x22
    OP_FXADDxs,      //0x23
    OP_FXADDxx,      //0x24
    OP_FXCMPdb,      //0x25
    OP_FXCMPsb,      //0x26
    OP_FXCMPxb,      //0x27
    OP_FXDIVdd,      //0x28
    OP_FXDIVdx,      //0x29
    OP_FXDIVss,      //0x2a
    OP_FXDIVsx,      //0x2b
    OP_FXDIVTdd,      //0x2c
    OP_FXDIVTdx,      //0x2d
    OP_FXDIVTss,      //0x2e
    OP_FXDIVTsx,      //0x2f
    OP_FXDIVxd,      //0x30
    OP_FXDIVxs,      //0x31
    OP_FXDIVxx,      //0x32
    OP_FXMULdd,      //0x33
    OP_FXMULdx,      //0x34
    OP_FXMULss,      //0x35
    OP_FXMULsx,      //0x36
    OP_FXMULxd,      //0x37
    OP_FXMULxs,      //0x38
    OP_FXMULxx,      //0x39
    OP_FXRSUBdd,      //0x3a
    OP_FXRSUBdx,      //0x3b
    OP_FXRSUBss,      //0x3c
    OP_FXRSUBsx,      //0x3d
    OP_FXSQRTIdx,      //0x3e
    OP_FXSQRTIsx,      //0x3f
    OP_FXSQRTIxx,      //0x40
    OP_FXSQRTTdx,      //0x41
    OP_FXSQRTTsx,      //0x42
    OP_FXSQRTTxx,      //0x43
    OP_FXSQRTUdx,      //0x44
    OP_FXSQRTUsx,      //0x45
    OP_FXSQRTUxx,      //0x46
    OP_FXSUBdd,      //0x47
    OP_FXSUBdx,      //0x48
    OP_FXSUBss,      //0x49
    OP_FXSUBsx,      //0x4a
    OP_FXSUBxd,      //0x4b
    OP_FXSUBxs,      //0x4c
    OP_FXSUBxx,      //0x4d
    OP_GETFd,      //0x4e
    OP_GETFs,      //0x4f
    OP_GETPL,      //0x50
    OP_LDB,      //0x51
    OP_LDCSB,      //0x52
    OP_LDCSD,      //0x53
    OP_LDCSH,      //0x54
    OP_LDCSW,      //0x55
    OP_LDD,      //0x56
    OP_LDDSB,      //0x57
    OP_LDDSD,      //0x58
    OP_LDDSH,      //0x59
    OP_LDDSW,      //0x5a
    OP_LDESB,      //0x5b
    OP_LDESD,      //0x5c
    OP_LDESH,      //0x5d
    OP_LDESW,      //0x5e
    OP_LDFSB,      //0x5f
    OP_LDFSD,      //0x60
    OP_LDFSH,      //0x61
    OP_LDFSW,      //0x62
    OP_LDGSB,      //0x63
    OP_LDGSD,      //0x64
    OP_LDGSH,      //0x65
    OP_LDGSW,      //0x66
    OP_LDH,      //0x67
    OP_LDSSB,      //0x68
    OP_LDSSD,      //0x69
    OP_LDSSH,      //0x6a
    OP_LDSSW,      //0x6b
    OP_LDW,      //0x6c
    OP_LZCNTd,      //0x6d
    OP_LZCNTs,      //0x6e
    OP_MERGEd,      //0x6f
    OP_MERGEs,      //0x70
    OP_MOVFI,      //0x71
    OP_MOVIF,      //0x72
    OP_MOVTd,      //0x73
    OP_MOVTs,      //0x74
    OP_MOVX,      //0x75
    OP_ORd,      //0x76
    OP_ORNd,      //0x77
    OP_ORNs,      //0x78
    OP_ORs,      //0x79
    OP_POPCNTd,      //0x7a
    OP_POPCNTs,      //0x7b
    OP_SARd,      //0x7c
    OP_SARs,      //0x7d
    OP_SCLd,      //0x7e
    OP_SCLs,      //0x7f
    OP_SCRd,      //0x80
    OP_SCRs,      //0x81
    OP_SDIVd,      //0x82
    OP_SDIVs,      //0x83
    OP_SDIVX,      //0x84
    OP_SHLd,      //0x85
    OP_SHLs,      //0x86
    OP_SHRd,      //0x87
    OP_SHRs,      //0x88
    OP_SMODX,      //0x89
    OP_STB,      //0x8a
    OP_STCSB,      //0x8b
    OP_STCSD,      //0x8c
    OP_STCSH,      //0x8d
    OP_STCSW,      //0x8e
    OP_STD,      //0x8f
    OP_STDSB,      //0x90
    OP_STDSD,      //0x91
    OP_STDSH,      //0x92
    OP_STDSW,      //0x93
    OP_STESB,      //0x94
    OP_STESD,      //0x95
    OP_STESH,      //0x96
    OP_STESW,      //0x97
    OP_STFSB,      //0x98
    OP_STFSD,      //0x99
    OP_STFSH,      //0x9a
    OP_STFSW,      //0x9b
    OP_STGSB,      //0x9c
    OP_STGSD,      //0x9d
    OP_STGSH,      //0x9e
    OP_STGSW,      //0x9f
    OP_STH,      //0xa0
    OP_STSSB,      //0xa1
    OP_STSSD,      //0xa2
    OP_STSSH,      //0xa3
    OP_STSSW,      //0xa4
    OP_STW,      //0xa5
    OP_SUBd,      //0xa6
    OP_SUBs,      //0xa7
    OP_SXT,      //0xa8
    OP_UDIVd,      //0xa9
    OP_UDIVs,      //0xaa
    OP_UDIVX,      //0xab
    OP_UMODX,      //0xac
    OP_VFSI,      //0xad
    OP_XORd,      //0xae
    OP_XORNd,      //0xaf
    OP_XORNs,      //0xb0
    OP_XORs,      //0xb1
    OP_NONE,      //0xb2
    OP_APTOAP,      //0xb3
    OP_APTOAPB,      //0xb4
    OP_CAST,      //0xb5
    OP_FCMPd,      //0xb6
    OP_FCMPdf,      //0xb7
    OP_FCMPs,      //0xb8
    OP_FCMPsf,      //0xb9
    OP_FDIVd,      //0xba
    OP_FDIVs,      //0xbb
    OP_FRCPs,      //0xbc
    OP_FRSQRTs,      //0xbd
    OP_FSQRTId,      //0xbe
    OP_FSQRTs,      //0xbf
    OP_FSQRTTd,      //0xc0
    OP_FXCMPdf,      //0xc1
    OP_FXCMPsf,      //0xc2
    OP_FXCMPxf,      //0xc3
    OP_GETSOD,      //0xc4
    OP_GETSP,      //0xc5
    OP_GETTAGd,      //0xc6
    OP_GETTAGs,      //0xc7
    OP_GETTC,      //0xc8
    OP_GETTD,      //0xc9
    OP_GETVA,      //0xca
    OP_INVTC,      //0xcb
    OP_LDAAB,      //0xcc
    OP_LDAAD,      //0xcd
    OP_LDAAH,      //0xce
    OP_LDAAQ,      //0xcf
    OP_LDAAW,      //0xd0
    OP_LDAPB,      //0xd1
    OP_LDAPD,      //0xd2
    OP_LDAPH,      //0xd3
    OP_LDAPQ,      //0xd4
    OP_LDAPW,      //0xd5
    OP_LDCSQ,      //0xd6
    OP_LDCUDB,      //0xd7
    OP_LDCUDD,      //0xd8
    OP_LDCUDH,      //0xd9
    OP_LDCUDQ,      //0xda
    OP_LDCUDW,      //0xdb
    OP_LDDSQ,      //0xdc
    OP_LDESQ,      //0xdd
    OP_LDFSQ,      //0xde
    OP_LDGDB,      //0xdf
    OP_LDGDD,      //0xe0
    OP_LDGDH,      //0xe1
    OP_LDGDQ,      //0xe2
    OP_LDGDW,      //0xe3
    OP_LDGSQ,      //0xe4
    OP_LDODPB,      //0xe5
    OP_LDODPD,      //0xe6
    OP_LDODPH,      //0xe7
    OP_LDODPQ,      //0xe8
    OP_LDODPW,      //0xe9
    OP_LDODRB,      //0xea
    OP_LDODRD,      //0xeb
    OP_LDODRH,      //0xec
    OP_LDODRQ,      //0xed
    OP_LDODRW,      //0xee
    OP_LDODWB,      //0xef
    OP_LDODWD,      //0xf0
    OP_LDODWH,      //0xf1
    OP_LDODWQ,      //0xf2
    OP_LDODWW,      //0xf3
    OP_LDRD,      //0xf4
    OP_LDSSQ,      //0xf5
    OP_MOVTq,      //0xf6
    OP_MULd,      //0xf7
    OP_MULs,      //0xf8
    OP_ODTOAP,      //0xf9
    OP_PACKSSHB,      //0xfa
    OP_PACKSSWH,      //0xfb
    OP_PACKUSHB,      //0xfc
    OP_PADDB,      //0xfd
    OP_PADDD,      //0xfe
    OP_PADDH,      //0xff
    OP_PADDSB,      //0x100
    OP_PADDSH,      //0x101
    OP_PADDUSB,      //0x102
    OP_PADDUSH,      //0x103
    OP_PADDW,      //0x104
    OP_PANDd,      //0x105
    OP_PANDNd,      //0x106
    OP_PAVGUSB,      //0x107
    OP_PAVGUSH,      //0x108
    OP_PCMPEQB,      //0x109
    OP_PCMPEQH,      //0x10a
    OP_PCMPEQW,      //0x10b
    OP_PCMPGTB,      //0x10c
    OP_PCMPGTH,      //0x10d
    OP_PCMPGTW,      //0x10e
    OP_PEXTRH,      //0x10f
    OP_PFADDd,      //0x110
    OP_PFADDs,      //0x111
    OP_PFCMPd,      //0x112
    OP_PFCMPs,      //0x113
    OP_PFDIVd,      //0x114
    OP_PFDIVs,      //0x115
    OP_PFMAXd,      //0x116
    OP_PFMAXs,      //0x117
    OP_PFMINd,      //0x118
    OP_PFMINs,      //0x119
    OP_PFMULd,      //0x11a
    OP_PFMULs,      //0x11b
    OP_PFSQRTs,      //0x11c
    OP_PFSQRTTd,      //0x11d
    OP_PFSUBd,      //0x11e
    OP_PFSUBs,      //0x11f
    OP_PINSH,      //0x120
    OP_PMADDH,      //0x121
    OP_PMAXSH,      //0x122
    OP_PMAXUB,      //0x123
    OP_PMINSH,      //0x124
    OP_PMINUB,      //0x125
    OP_PMOVMSKB,      //0x126
    OP_PMOVMSKPD,      //0x127
    OP_PMOVMSKPS,      //0x128
    OP_PMULHH,      //0x129
    OP_PMULHUH,      //0x12a
    OP_PMULLH,      //0x12b
    OP_PMULUBHH,      //0x12c
    OP_PORd,      //0x12d
    OP_PSADBW,      //0x12e
    OP_PSHUFH,      //0x12f
    OP_PSHUFW,      //0x130
    OP_PSLLD,      //0x131
    OP_PSLLH,      //0x132
    OP_PSLLQH,      //0x133
    OP_PSLLQL,      //0x134
    OP_PSLLW,      //0x135
    OP_PSRAH,      //0x136
    OP_PSRAW,      //0x137
    OP_PSRLD,      //0x138
    OP_PSRLH,      //0x139
    OP_PSRLQH,      //0x13a
    OP_PSRLQL,      //0x13b
    OP_PSRLW,      //0x13c
    OP_PSUBB,      //0x13d
    OP_PSUBD,      //0x13e
    OP_PSUBH,      //0x13f
    OP_PSUBSB,      //0x140
    OP_PSUBSH,      //0x141
    OP_PSUBUSB,      //0x142
    OP_PSUBUSH,      //0x143
    OP_PSUBW,      //0x144
    OP_PUNPCKHBH,      //0x145
    OP_PUNPCKHHW,      //0x146
    OP_PUNPCKHWD,      //0x147
    OP_PUNPCKLBH,      //0x148
    OP_PUNPCKLHW,      //0x149
    OP_PUNPCKLWD,      //0x14a
    OP_PUTTAGd,      //0x14b
    OP_PUTTAGs,      //0x14c
    OP_PUTTC,      //0x14d
    OP_PXORd,      //0x14e
    OP_RRd,      //0x14f
    OP_RRs,      //0x150
    OP_RWd,      //0x151
    OP_RWs,      //0x152
    OP_SMULX,      //0x153
    OP_STAAB,      //0x154
    OP_STAAD,      //0x155
    OP_STAAH,      //0x156
    OP_STAAQ,      //0x157
    OP_STAAW,      //0x158
    OP_STAPB,      //0x159
    OP_STAPD,      //0x15a
    OP_STAPH,      //0x15b
    OP_STAPQ,      //0x15c
    OP_STAPW,      //0x15d
    OP_STCSQ,      //0x15e
    OP_STDSQ,      //0x15f
    OP_STESQ,      //0x160
    OP_STFSQ,      //0x161
    OP_STGDB,      //0x162
    OP_STGDD,      //0x163
    OP_STGDH,      //0x164
    OP_STGDQ,      //0x165
    OP_STGDW,      //0x166
    OP_STGSQ,      //0x167
    OP_STODPB,      //0x168
    OP_STODPD,      //0x169
    OP_STODPH,      //0x16a
    OP_STODPQ,      //0x16b
    OP_STODPW,      //0x16c
    OP_STODRB,      //0x16d
    OP_STODRD,      //0x16e
    OP_STODRH,      //0x16f
    OP_STODRQ,      //0x170
    OP_STODRW,      //0x171
    OP_STODWB,      //0x172
    OP_STODWD,      //0x173
    OP_STODWH,      //0x174
    OP_STODWQ,      //0x175
    OP_STODWW,      //0x176
    OP_STRD,      //0x177
    OP_STSSQ,      //0x178
    OP_TDTOMP,      //0x179
    OP_UMULX,      //0x17a
    OP_FDTOIFD,    //0x17b
    OP_SMULH,      //0x17c
    OP_UMULH,      //0x17d
#ifdef __cplusplus
    OP_LAST,      //0x17e
    // Pseudo ops
    OP_CT,
    OP_CALL,
    OP_RLP,
    OP_ELP,
    OP_RETURN,
    OP_DISP,
    OP_PSEUDO_LAST
#endif // __cplusplus
} E2kOpCode;

typedef enum {
    LD_MAS_NORMAL = 0x0,
    LD_MAS_SEMI_SPEC = 0x0,
    LD_MAS_DIS_TRANSL = 0x1,
    LD_MAS_SPEC_2ND = 0x1,
    LD_MAS_CHECK = 0x2,
    LD_MAS_CHECK_UNLOCK = 0x3,
    LD_MAS_SPEC = 0x3,
    LD_MAS_FILL = 0x4,
    LD_MAS_LOCK_CHECK = 0x4,
    LD_MAS_TRAP_S = 0x5,
    LD_MAS_TRAP_LS = 0x6,
    LD_MAS_LOCK_WAIT = 0x7,
    LD_MAS_SPEC_LOCK_CHECK = 0x7
} LdMasTypes;

typedef enum {
    ST_MAS_NORMAL = 0x0,
    ST_MAS_SEMI_SPEC = 0x0,
    ST_MAS_DIS_TRANSL = 0x1,
    ST_MAS_SPEC_2ND = 0x1,
    ST_MAS_WAIT_UNLOCK = 0x2,
    ST_MAS_STORE_UNLOCK = 0x3,
    ST_MAS_NO_CHECK = 0x4,
    ST_MAS_UNLOCK = 0x5,
    ST_MAS_IO_OP = 0x6,
    ST_MAS_SPECIAL = 0x7
} StMasTypes;

typedef enum {
    WAIT_ALL_C    = BIT(0),
    WAIT_ALL_E    = BIT(1),
    WAIT_ST_C     = BIT(2),
    WAIT_LD_C     = BIT(3),
    WAIT_FL_C     = BIT(4),
    WAIT_MA_C     = BIT(5),
    WAIT_TRAP     = BIT(6),
    WAIT_ALL      = (WAIT_ALL_C | WAIT_ALL_E | WAIT_ST_C | WAIT_LD_C | WAIT_FL_C | WAIT_MA_C | WAIT_TRAP)
} WaitInstType;

#ifdef __cplusplus

enum OutRegLantencyType : uint8_t;
enum InRegLantencyType : uint8_t;

typedef struct ALUInfo {
    const char* mnem;
    ElbrusOpcode opcode;
    ExecUnitMask alu_mask;
    OutRegLantencyType out_latency;
    InRegLantencyType in_latency;
} ALUInfo;

#endif // __cplusplus

NAMESPACE_ASM_E2K_END

#endif // ASM_E2K_OPCODES_E2K_
