#ifndef TANGO_TANGO_H
#define TANGO_TANGO_H

#include "tango/ir.hpp"
#include "tango/compilation_session.hpp"


namespace tango {

void setSharedAllocator(TangoObj::Allocator *a);

void ssa(Ir *ir);

void fastCopyCoalescing(Ir *ir);

void blocksMerge(Ir *ir);

void loopBuild(Ir *ir);

void dynchecksGenerator(Ir *ir);

void loopTransformation(Ir *ir);

void loopInvariants(Ir *ir);

void ifConversion(Ir *ir);

void coalescer(Ir *ir);

void copyPropagation(Ir *ir);

void blockCloner(Ir *ir);

void softwarePipeliner(Ir *ir);

template<bool SORT_BLOCK = true>
void codeLayouter(Ir *ir);

template<bool BUNDLE = true>
void deadCodeEliminator(Ir *ir);

void ssaDeadCodeEliminator(Ir *ir);

void verifier(Ir *ir);

void peephole(Ir *ir);

void instScheduler(Ir *ir);

void coldPathMerge(Ir *ir);

template<PeepHoleMode PH_MODE>
void doMacroInsts(Ir *ir);

void registerAllocator(Ir *ir);

void liveAnalyser(Ir *ir);

void spillGen(Ir *ir);

void prologGenerator(Ir *ir);

void stateFlusher(Ir *ir);

void sxtCorrector(Ir *ir);

void landP(Ir *ir);

void emitter(Ir *ir, CodeBuffer *codeBuffer);

//----------------------------------------------------------------------//

void startSession(Ir*, size_t, CompType, CompMode, int, const char*, const char*, Options* = nullptr);

void closeSession();

void exitSession(const char*);

void initPassManager();

void resetLogOutputType();

uint32_t generateId();

void printAsm(const char* = nullptr);

void printCfgDot(const char*);

void printBlocks(const char* = nullptr);

void printLoops(const char* = nullptr);

//----------------------------------------------------------------------//

#ifdef TANGO_LOG_ENABLED

void setLocalLogger(LocalLogger*);

bool verifyTangoLog(const char*);

void initGlobalLogger();

void initOptions(Options&, const char*);

#endif

}
#endif //TANGO_TANGO_H
