#ifndef TANGO_BLOCK_CLONER_HPP
#define TANGO_BLOCK_CLONER_HPP

#include "tango/compile_phase.hpp"
#include "tango/clone_helper.hpp"

namespace tango {

//----------------------------------------------------------------------------//

class BlockCloner : public CompilePhase {
public:
  BlockCloner(Ir* ir):
    CompilePhase(ir, PhaseName::BCL), block_clone_helper(ir) {
    clone_limit = ir->blocks.size() / 2;
  }
  void   changeLabel(Block* src, Block* oldBlock, Block* newBlock);

protected:
  virtual void run_internal();

private:
  void   createClones(BlockList& blocks);
  void   makeSuperblockLongerFrom(Block* block);
  bool   wantClone(const Block* block) const;

  void   cloneSuccs(Block*, Block*);
  Block* findNextBlockForTrace(Block* block);
  Block* cutBlockByPred(Block*, Block*, float prob);
  bool   canAddToTrace(Block* block);
  bool   canAddToTraceAfter(Block* block);
  bool   canBeLayout(Block* dst, Block* src);

private:
  BlockSet heads; // Head can be added to another trace.
  BlockSet trace; // Block inside current trace.
  BlockCloneHelper block_clone_helper;
  int clone_limit;
};

//----------------------------------------------------------------------------//

}      // namespace tango
#endif // TANGO_BLOCK_CLONER_HPP
