#ifndef TANGO_CFG_HPP
#define TANGO_CFG_HPP

#include "tango/ir.hpp"

namespace tango {

typedef arena_map <Opnd*, Opnd*>::type OpndReplace;
typedef arena_map <Block*, Block*>::type BlockReplace;

void copyTouchUses(const Inst* inst, Inst* new_inst, OpndReplace& replaced);

namespace cfg   {

void   removeBlock(Ir* ir, Block* pred, Block* block, Block* succ);

// Is empty when has no insts but branch and a single succ and pred
bool   isEmptyBlock(Block* block);

Block* getSuccessorByIndex(Block* block, uint num, bool skip_empty = false);
bool   hasDirectPath (Block* head, Block* tail);
bool   isFalsePath(Inst* inst, Block* tail);
bool   isTruePath(Inst* inst, Block* tail);
Inst*  addBranch(Ir* ir, Block* from, Block* to);
void   changeLabel(Block* src, Block* old_block, Block* new_block);

// Adds block after pred and before succ changing label
void   addBlockBetween(Block* new_block, Block* pred, Block* succ);
// Changes all loop preds to new_block
void   addBlockBefore(Block* initial_block, Block* new_block);
void   addBlockAfter(Block* initial_block, Block* new_block); // Returns next successor

// Splits block into two sequentional: [begin, splitter_inst)->[splitter_inst, end)
// returns new block
Block* splitBlock(Ir* ir, Block* block, Inst* splitter_inst);

// Changes labels and succs for uncheduled block
void   redirectJump(Block* block, Block* old_block, Block* new_block);

// Returns block successor if block has only one or null otherwise
Block* getDirectSucc(Block* block);

// Fills jump predicate and predicate inverse for jump
void   getJumpPredicate(Block* from, Block* to, Opnd** pred);

Block* getLastConnector(Block* block);

void mergeGoto(Ir* ir, Block *pred, Block *bc, Block *succ);
void replaceSucc(Block* where, Block* what, Block* with);
void mergeRethrow(Ir* ir, Block *pred, Block *bc, Block *succ);
void mergeSwitch(Block *pred, Block *bc, Block *succ);
void mergeCondBranch(Block *pred, Block *bc, Block *succ);

} // namespace cfg
} // namespace tango

#endif // TANGO CFG HPP
