#ifndef TANGO_CLONE_HELPER_HPP
#define TANGO_CLONE_HELPER_HPP

#include "tango/ir.hpp"
#include "tango/live_analyser.hpp"

namespace tango {

//----------------------------------------------------------------------------//

class CloneHelper : public TangoObj {
public:
  CloneHelper(Ir* ir) : ir(ir) {}

  virtual bool canClone(const Inst* inst) const;
  bool         canCopy(const Block* block);
  Inst*        clone(const Inst* inst) const;
  Inst*        copy(const Inst* inst) const;
  Block*       createCompBlock(Block* target, InstList& insts);
  Block*       createRbranchCompBlock(InstList& insts, Block* original);
  void         cloneInsts(const InstList& inst, Block* dst);
  void         copyInsts(const InstList& inst, const Block* src, Block* dst);

  void         setLiveOutSet(const LiveSet& live_set);
  void         deleteLiveOutSet();
  void         copyBlock(const Block* src, Block* dst);
  Loop*        copyLoop(Loop* l, float prob);

protected:
  bool         mustRestore(Inst* inst) const;

  Ir*            ir;  
  const LiveSet* live_set;
};

//----------------------------------------------------------------------------//

class BlockCloneHelper : public CloneHelper {
public:
  BlockCloneHelper(Ir* ir) : CloneHelper(ir) {}
  void cloneBlock(const Block* src, Block* dst);
  bool canClone(const Inst* inst) const;
  bool canClone(const Block* block);
};

}      // namespace tango
#endif // TANGO_CLONE_HELPER_HPP
