#ifndef TANGO_COPY_PROPOGATION_HPP
#define TANGO_COPY_PROPOGATION_HPP

#include "tango/compile_phase.hpp"

namespace tango {

#define USE_GLOBAL_COPY_PROPAGATION 0
#if USE_GLOBAL_COPY_PROPAGATION
#  define GLOBAL_COPY_PROP(a) a
#else
#  define GLOBAL_COPY_PROP(a)
#endif // USE_GLOBAL_COPY_PROPAGATION

// def <- use
class AssignmentPair : public TangoObj {
public:
  AssignmentPair(Opnd *_def, Opnd *_use): def(_def), use(_use) {}
  bool isValid() const {return def != NULL && use != NULL;}
  Opnd *def;
  Opnd *use;

  bool operator< (const AssignmentPair& r) const {
    if ( def != r.def )
      return def < r.def;
    else
      return use < r.use;
  }
};

struct AssignmentSet : public TangoObj {
  AssignmentSet(Ir* ir, bool init) : AssignmentSet(ir->max_opnd_id, ir->allocator(), init) { }

  AssignmentSet(int max_opnd_id, Allocator* alloc, bool init);

  AssignmentSet(const AssignmentSet &a);

  void initialize(bool fill);

  void check_initialize() { if(!_init) initialize(true); }

  int size() { return _size; }

  void insert(Opnd* def, Opnd *use);

  void insert(AssignmentPair a) { insert(a.def, a.use); }

  void intersection(AssignmentSet &set) {
    // TODO:: implement
  }

  void removeConflicts(Opnd *def);

  Opnd* findReplacement(Opnd *use);

  AssignmentSet & operator= ( const AssignmentSet & );

  void reset();

  Allocator* alloc;
  int max_opnd_id;
  Opnd** opnds;
  int* def2use;
  int* use2def;
  int _size;
  bool _init;
};

typedef arena_vector<AssignmentSet >::type AssignmentSetVector;

class CopyPropagation : public CompilePhase {
public:
  CopyPropagation(Ir* ir) : CompilePhase(ir, PhaseName::CPN), tmp(ir, false) {}
  void          init();
  void          localUseCopyPropogation(AssignmentSet& acp, OpndVector& uses, int32_t src2_pos);
  void          localTouchUseCopyPropogation(AssignmentSet& acp, OpndVector& uses);
#if USE_GLOBAL_COPY_PROPAGATION
  AssignmentSet
#else
  void
#endif
  localCopyPropagation(Block* block, AssignmentSet& acp);
  void          globalCopyPropagation();
  AssignmentSet computeCPIn(Block *block);
  bool          isAliveAssignment(Inst *inst);

protected:
  virtual void run_internal();

public:
  AssignmentSetVector copy;
  AssignmentSetVector kill;
  AssignmentSetVector out;
  AssignmentSetVector in;
  bool                dfChanges;

private:
  Opnd*          findReplacement(Opnd *val, AssignmentSet &acp);
  void           findLocalConflict(Opnd *val, AssignmentSet &acp);
  void           findGlobalConflict(Opnd *val, AssignmentSet &acp);
  AssignmentPair killedAssignment(Opnd *def);

  int            countMoves();

  AssignmentSet  tmp;
};

}      // namespace tango
#endif // TANGO_COPY_PROPOGATION_HPP
