#ifndef  TANGO_UTILITIES_DISJOINTED_SETS_HPP
#define  TANGO_UTILITIES_DISJOINTED_SETS_HPP

#include <tango/tango_allocator.hpp>
#include "tango/tango_inst.hpp"

namespace tango {

template <typename T> class Cell : public TangoObj {
public:
  Cell(T elem): _elem(elem), _ptr(this), _weight(0) {}

  T        val() const    { return _elem; }
  Cell<T>* ptr() const    { return _ptr; }
  int      weight() const { return _weight; }

  void set_parent(Cell<T>* ptr) { _ptr = ptr; }
  void set_weight(int w)        { _weight = w; }

private:
  T        _elem;
  Cell<T>* _ptr;
  int      _weight;
};


template <typename T> class DisjointedSets : public TangoObj {
public:
  DisjointedSets() {}

  Cell<T>* makeset(T x, Allocator* alloc);
  Cell<T>* find(Cell<T>* x);
  Cell<T>* find(T x);
  Cell<T>* unite(Cell<T>* x, Cell<T>* y);
  bool     intersect(Cell<T>* x, Cell<T>* y);
  void     separate(Cell<T>* x);

  typename arena_vector<Cell<T>*>::type _sets;
};




} // namespace tango

#endif // TANGO_UTILITIES_DISJOINTED_SETS_HPP
