#ifndef TANGO_DOMINATORS_HPP
#define TANGO_DOMINATORS_HPP

#include "tango/ir.hpp"

namespace tango {

// Dominator tree is built using the algorithm from
// Keith D. Cooper, Timothy J. Harvey, and Ken Kennedy.
// A Simple, Fast Dominance Algorithm.

class DomTree : public TangoObj {
public:
  DomTree(Ir* ir): _ir(ir) {}
  void build();
  bool dominates(Block* dominator, Block* block) const;

  Block* idom(Block* b) const {
    return _doms[b->rpo];
  }

private:
  Block* getFirstProcessedPred(Block* block, BlockList::iterator &it);
  Block* intersect(Block* b1, Block* b2);

public:
  // Combined dominator sets for all blocks while iterating
  // and the dominator tree after the algorythm is done.
  // _doms[i] contains the immediate dominator of the block with rpo i.
  BlockVector _doms;
  Ir* _ir;
};


}

#endif // TANGO DOMINATORS HPP
