#ifndef TANGO_EMITTER_HPP
#define TANGO_EMITTER_HPP

#include "tango/compile_phase.hpp"
#include "tango/encoder.hpp"

namespace tango {

//----------------------------------------------------------------------------//

class Disp;
typedef arena_list <Disp*>::type DispList;

class Switch;
typedef arena_list <Switch*>::type SwitchList;

//----------------------------------------------------------------------------//

class Disp : public TangoObj {
public:
  Disp(Block* block, uint64_t bundle_addr) : block(block), bundle_addr(bundle_addr) {}
  Block*      block;
  uint64_t    bundle_addr;
  int32_t     c0_offset;
};

//----------------------------------------------------------------------------//

class Switch : public TangoObj {
public:
  Switch(ImmOpnd* jumpTable, OpndVector::iterator cases_begin, OpndVector::iterator cases_end) :
    jumpTable(jumpTable), cases(cases_begin, cases_end) {}

  ImmOpnd*     jumpTable;
  OpndVector   cases;
};

//----------------------------------------------------------------------------//

class Emitter : public CompilePhase {
public:
  Emitter(Ir*, /*uint16_t**/CodeBuffer *);
  void doEmit(Bundle*);
  void doEmit(Inst*);

protected:
  ElbrusOpcode opc(Opnd*, ElbrusOpcode, ElbrusOpcode);
  ElbrusOpcode comb_opc(Opnd*, ElbrusOpcode, ElbrusOpcode);
  ElbrusOpcode movOpc(Opnd*, Opnd*);
  Opnd*        sxtOpc(Opnd*, bool);
  void         emitCmp(ElbrusOpcode, ElbrusOpcode, ElbrusOpcode, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, bool, bool);
  void         emitCmpM(ElbrusOpcode, ElbrusOpcode, ElbrusOpcode, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, bool, bool);
  void         emitItoF(uint32_t, Opnd*, Opnd*, Opnd*, bool, bool);
  void         emitFtoI(uint32_t, Opnd*, Opnd*, Opnd*, bool, bool);
  void         emitFtoD(uint32_t, Opnd*, Opnd*, Opnd*, bool, bool);
  void         emitDtoF(uint32_t, Opnd*, Opnd*, Opnd*, bool, bool);
  void         emitFtoX(uint32_t, Opnd*, Opnd*, Opnd*, bool, bool);
  void         emitDtoX(uint32_t, Opnd*, Opnd*, Opnd*, bool, bool);
  void         emitSwitch(uint32_t, Opnd*, OpndVector, bool);
  void         emitExSwitch(uint32_t, Opnd*, OpndVector, bool);
  void         newDisp(Inst*, uint32_t = 1);
  void         patchDisps();
  void         patchSwitches();
  void         emitFxmulx(uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, bool, bool);
  void         emitFxmuld(uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, bool, bool);
  void         emitFxadd(uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, bool, bool);
  void         emitBranchPrep(Opnd*, ElbrusOpcode, Opnd*);
  void         emitMovtd(int32_t, Opnd*, Opnd*, Opnd*, bool, bool);
  void				 emitAbn(Opnd*, Opnd*, Opnd*, Opnd*);
  void				 emitBap(Opnd*, Opnd*);

  void         emitLoad(ElbrusOpcode, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, uint32_t, bool, bool);
  void         emitStore(ElbrusOpcode, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, uint32_t, bool, bool);

  void         emitArrayRead(ElbrusOpcode, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, Opnd*, Opnd*, bool, uint32_t, bool, bool);
  void         emitArrayRegRead(ElbrusOpcode, uint32_t, Opnd*, Opnd*, Opnd*);

  void         emitArrayStore(ElbrusOpcode, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, Opnd*, Opnd*, bool, uint32_t, bool, bool);
  void         emitArrayRegWrite(ElbrusOpcode, uint32_t, Opnd*, Opnd*, Opnd*);

  void         emitLdisp(Opnd*, ElbrusOpcode, const OpndVector&, const OpndVector&);
  void         emitMova(ElbrusOpcode, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*);

  void         ibranchFix();
  void         memoryIntersectFix();

  void         doCopy(Inst*);

protected:
  virtual void run_internal();

protected:
  Encoder      e;
  DispList     disps;      // bundles containing references to blocks
  Disp*        disp;       // disp info of bundle being emitted
  RelocLDisp*  ldisp_info; // used to patch ldisp address after emitting bundle
  SwitchList   switches;

  Block*       cur_block;

  //uint16_t*    code;
  CodeBuffer*  codeBuffer;

  Opnd*        imm0;
  Opnd*        imm1;
  Opnd*        imm3;
};

//----------------------------------------------------------------------------//

}      // namespace tango
#endif // TANGO_EMITTER_HPP
