#ifndef TANGO_ENCODER_H
#define TANGO_ENCODER_H

#include "tango/ir.hpp"
#include "tango/opcodes.hpp"

namespace tango {

class Lts;

//----------------------------------------------------------------------------//

class Lts;
typedef arena_vector<Lts>::type LtsVector;

//----------------------------------------------------------------------------//

enum LtsType {
  LTS_SPECIAL,
  LTS_AAOPCE1,
  LTS_16,
  LTS_32,
  LTS_64,
  LTS_LAST
};

//----------------------------------------------------------------------------//

class BitWriter : public TangoObj {
public:
  BitWriter(CodeBuffer* codeBuffer) : codeBuffer(codeBuffer), buff(0), flag(false) {}
  void            init();
  void            flush();
  void            slbl(uint32_t);
  void            half(uint16_t);

public:
  CodeBuffer* codeBuffer;
  uint32_t        buff;
  bool            flag;
};

//----------------------------------------------------------------------------//

class Lts : public TangoObj {
public:
  Lts(LtsType, uint64_t, int32_t, Reloc*);

public:
  LtsType     type;       // 16, 32, 64 or special LTS0
  uint64_t    val;        // lts value
  int32_t     als_idx;    // index in als vector
  int32_t     size;       // lts size in bytes
  Reloc *     reloc;
};

//----------------------------------------------------------------------------//

class Encoder : public TangoObj {
public:
  Encoder(CodeBuffer* codeBuffer, DebugInfoTransfer* debug_info, CPUInfo* cpuinfo = nullptr)
    : code_buffer(codeBuffer), bw(codeBuffer), relocs(debug_info), cpuinfo(cpuinfo) {}

  // bundle encoding
  void             startBundle();
  void             closeBundle();

  // instructions having no format
  void             nop(Opnd*);
  void             ibranch(Opnd*, Opnd*, bool, uint32_t=0, uint32_t clp =0);
  void             mcmpibranch(Opnd *px, Opnd*, Opnd*, Opnd*, Opnd*, Opnd *def, bool, uint32_t, uint32_t als, InstCode opcode, bool);
  void             movibranch(ElbrusOpcode, Opnd*, Opnd*, Opnd*, Opnd*, Opnd*, bool, uint32_t, uint32_t, bool);
  void             ct(Opnd*, Opnd*, bool=false, uint32_t=0);
  void             call(Opnd*, Opnd*, Opnd*, bool);
  void             merge(uint32_t, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, Opnd*, bool=false, bool=false);
  void             ld_mas(Opnd*, Opnd*, Opnd*, Opnd*, Opnd*);
  void             std_unlock_rbranch(Opnd*, Opnd*, Opnd*, Opnd*, Opnd*);
  void             andp(Opnd*, Opnd*, Opnd*, uint32_t=0, uint32_t=0, bool=true);
  void             mlandp(Opnd*, Opnd*, Opnd*, Opnd*, Opnd*, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t);
  void             movep(Opnd*, Opnd*, Opnd*, uint32_t=0, uint32_t=0);
  void             fapb(Opnd*, Opnd*, Opnd*);

  // operations encoding
  void             ALOPF1(uint32_t, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, bool=false/*inverse*/, bool = false/*speculative*/);
  void             ALOPF2(uint32_t, uint32_t, uint32_t, Opnd*, Opnd*, Opnd*, bool=false, bool = false);
  void             ALOPF3(uint32_t, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, bool=false, bool = false);
  void             ALOPF7(uint32_t, uint32_t, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, bool, bool = false);
  void             ALOPF8(uint32_t, uint32_t, uint32_t, Opnd*, Opnd*, Opnd*, bool, bool = false);
  void             ALOPF10(uint32_t, uint32_t, uint32_t, uint32_t, Opnd*, Opnd*, bool=false, bool=false);
  void             ALOPF11(uint32_t, uint32_t, uint32_t, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, bool=false, bool = false);
  void             ALOPF12(uint32_t, uint32_t, uint32_t, Opnd*, Opnd*, Opnd*, bool=false, bool = false);
  void             ALOPF12(uint32_t, uint32_t, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, bool=false, bool = false);
  void             ALOPF13(uint32_t, uint32_t, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, bool=false, bool = false);
  void             ALOPF15(uint32_t, uint32_t, uint32_t, Opnd*, Opnd*, uint32_t, bool=false, bool = false);
  void             ALOPF16(uint32_t, uint32_t, uint32_t, Opnd*, uint32_t, Opnd*, bool=false, bool = false);
  void             ALOPF17(uint32_t, uint32_t, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, bool=false, bool = false);
  void             ALOPF19(uint32_t, uint32_t, uint32_t, uint32_t, Opnd*, Opnd*, bool=false, bool=false);
  void             ALOPF21(uint32_t, uint32_t, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, Opnd*, bool=false, bool = false);
  void             ALOPF22(uint32_t, uint32_t, uint32_t, Opnd*, uint32_t, Opnd*, Opnd*, Opnd*, bool=false, bool = false);
  void             COPF2(Opnd*, uint32_t, Opnd*);
  void             COPF4(uint32_t, Opnd*);
  void             CDOPF1(uint32_t, uint32_t, Opnd*, bool=false);
  void             PLOPF0(uint32_t, uint32_t, Opnd*, uint32_t, Opnd*, uint32_t, Opnd*);
  void             PLOPF1(uint32_t, uint32_t, Opnd*, uint32_t, Opnd*, uint32_t, Opnd*);
  void             PLOPF012(uint32_t, uint32_t, Opnd*, uint32_t, Opnd*, uint32_t, Opnd*, uint32_t, Opnd*, uint32_t, uint32_t, Opnd*, Opnd* = nullptr, Opnd* = nullptr);
  void             AAOPF1(uint32_t, uint32_t, Opnd*, Opnd*, Opnd*, Opnd*);

  // extended operations encoding
  void             ALOPF2ext(uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, Opnd*, Opnd*, Opnd*, bool, bool);
  void             COPF4ext(uint32_t, Opnd*, Opnd*, Opnd*, Opnd*, Opnd*, Opnd*);

public:
  // syllables encoding
  void             SF1(uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t);
  void             ALF1_8(int32_t, uint32_t, uint32_t, uint32_t, uint32_t, bool);
  void             ALF6(int32_t, uint32_t, uint32_t, uint32_t, bool);
  void             ALF9_10(int32_t, uint32_t, uint32_t, uint32_t, bool);
  void             ALEF1_2(int32_t, uint32_t, uint32_t);
  void             C0F2(uint32_t, uint32_t, uint32_t);
  void             C1F1(uint32_t, uint32_t);
  void             PL0F1(uint32_t, uint32_t, uint32_t);
  void             PL1F1(uint32_t, uint32_t, uint32_t);
  void             PL2F1(uint32_t);
  void             AA2F1(uint32_t, bool, uint32_t, uint32_t, uint32_t, bool);
  void             AA0_1F1(uint32_t, uint32_t);
  
  // AA2F1, AA0F1

  // operands encoding
  uint32_t         SRC1(Opnd*);
  uint32_t         SRC2(Opnd*, int32_t);
  uint32_t         SRC3(Opnd*);
  uint32_t         DST(Opnd*);
  uint32_t         DST2(uint32_t, Opnd*);
  uint32_t         ELP(Opnd*);
  uint32_t         LTS0_SPECIAL(uint32_t, uint32_t, uint32_t, uint32_t, uint32_t);
  uint32_t         AA_OPCE1(int32_t, Opnd*, Opnd*, Opnd*, Opnd*, bool);
  uint32_t         AA_OPCE1_Special(Opnd*);

  // literal slogs support
  bool             newLts(LtsType, uint64_t, int32_t=-1, Reloc * = NULL);
  void             setLtsForAls(uint32_t, uint32_t);
  void             prepareLts();
  void             encodeLts();

  // helpers
  uint32_t         getPr(Opnd*);
  uint32_t         getCondType(Opnd*, bool, uint32_t);
  void             prepareHS();
  void             prepareCD();
  void             prepareLen();

  ElbrusOpcode     memReadOpc(DataType type) const;
  ElbrusOpcode     memWriteOpc(DataType type) const;
public:
  // header fields
  uint32_t         hs_al;
  uint32_t         hs_ale;
  uint32_t         hs_pl;
  uint32_t         hs_cd_;  // num of rlp operations
  uint32_t         hs_c;
  uint32_t         hs_sw;
  uint32_t         hs_s;
  uint32_t         hs_x;
  uint32_t         hs_lm;
  uint32_t         hs_nop;
  uint32_t         hs_lng_; // in half words
  uint32_t         hs_mdl_; // in half words
  
  // stubs syllable fields
  uint32_t         ss_aa;

  // bundle fields
  uint32_t         hs;
  uint32_t         ss;
  uint32_t         als[6];
  uint32_t         cs[2];
  uint32_t         pad;
  uint16_t         ale[6];
  uint16_t         aas[6];
  LtsVector        ltss;
  uint32_t         pls[3];
  uint16_t         cds[6];

  int32_t          lts_len;       // len of lts area in bytes
  int32_t          lts16_num;     // num allocated slots for lts16 (can't be > 4)
  int32_t          c0_offset;     // c0 syllable offset in bundle
  CodeBuffer*      code_buffer;
  BitWriter        bw;

  DebugInfoTransfer*   relocs;
  CPUInfo*             cpuinfo;
};

} // namespace tango
#endif // TANGO_ENCODER_H
