#ifndef TANGO_SSA_FCC_H
#define TANGO_SSA_FCC_H

#include "tango/compile_phase.hpp"
#include "tango/disjointed_sets.hpp"

namespace tango {

class FastCopyCoalescing;

struct ForestNode : public TangoObj {
  ForestNode(Block* b) : block(b) {}
  Block* block;
  arena_vector<ForestNode*>::type succs;
};

class FastCopyCoalescing : public CompilePhase {
public:
  FastCopyCoalescing(Ir* ir/*, Selector* sel*/): CompilePhase(ir, PhaseName::FCC) /*, _sel(sel)*/ {}
  void  determineDefSites();
  void  coalesce();
  void  predictTraces();
  void  disjoinPhis();
  void  walkForest(ForestNode* fn);
  void  checkAndBreak(ForestNode* p, ForestNode* c);
  void  breakPhi(Opnd* p, Inst* phi, uint idx);
  void  breakPhiCompletely(Inst* phi);
  Opnd* checkLocal(Opnd *p, Opnd* c, Block* b);
  void  checkBlockDefs(ForestNode *p);
  void  insertCopies();
  void  validateCopies();
  Opnd* chooseCandidate(Opnd* op1, Opnd* op2);
  void  createSets();
  void  separate(Opnd* n);
  virtual void  patch();
  void  removePhis();

  Block* getNextHead(BlockList& blocks, Block* src);
  Block* getLayout(Block* block);
  Block* getTrg(Inst* ct_cond, int32_t trg_pos);
  Block* getGotoLayout(Block* block);
  Block* getCondLayout(Block* block, Inst* ct_cond);

  Cell<Opnd*>* initCell(Opnd* op);

protected:
  virtual void run_internal();

protected:
  arena_vector<arena_vector<Inst*>::type>::type waiting;    // Order of copy insertion
  arena_map<Block*, OpndVector>::type           block_phis; // List of phi-related defs for each block
  arena_map<Opnd*, int32_t>::type               traces;     // Every block has an associated trace in numerical representation
  arena_vector<int>::type                       phi_opnds;  // Temporary list for storing id's of phi-related operands
  arena_vector<Cell<Opnd*>*>::type              cells;      // Mapping of operands on cells
  DisjointedSets<Opnd*>                         name_sets;  // Operands to coalesce
  arena_map<Opnd*, InstVector>::type            opnd2phis;  // For each operand list of phi-nodes in which it's found
  int32_t                                       max_trace;
};

}
#endif // TANGO_SSA_FCC_H
