#ifndef TANGO_IF_CONVERSION_HPP
#define TANGO_IF_CONVERSION_HPP

#include "tango/compile_phase.hpp"
#include "tango/block_cloner.hpp"
#include "tango/cfg.hpp"
#include <algorithm>

namespace tango {

//----------------------------------------------------------------------------//

class IfConversion : public CompilePhase {
public:
  IfConversion(Ir* ir) : CompilePhase(ir, PhaseName::IFC), clone_helper(ir), branch1_size(0), branch2_size() {}

protected:
  bool  processIf(Block *btop, Block *branch1, Block *branch2, Block *btail);

  // a) case from figure
  bool  processTwoBranches(Block* btop, Block* branch1, Block* branch2);

  // b) case from figure
  bool  processBtailBranch(Block* btop, Block* branch, Block* btail);

  bool  validBranch(Block *branch);
  bool  validTop(Block *btop, Block *branch1, Block *branch2);
  
  virtual void run_internal();
  
private:
  void  processBtail(InstList& res_insts, Block* btail, Block* true_b,
                     Block* false_b, RegOpnd* ct_pred) const;

  bool  needsPredicate(const Inst* inst) const;
private:
  CloneHelper         clone_helper;

  int                 branch1_size;
  int                 branch2_size;
};

//----------------------------------------------------------------------------//

class LoopIfConversion : public TangoObj {
public:
  LoopIfConversion(Ir* ir) : _ir(ir) {}

  void   processLoop(Loop* loop);
  bool   validLoop(Loop* loop) const;
protected:
  using  Pred = std::pair<int /* id */, bool /* inverted */>;
  using  PredMap = arena_map<std::pair<Pred, Pred>, Opnd*>::type;

  void   reset();

  // Find next block in loop (B[i + 1] if cur == B[i])
  Block* getNext(Block* block) const;

  // Returns inst andp pred1, pred2 -> def ? condition_pred
  // Uses inverses determined by preds->isInversed()
  // Adds combined predicate to _preds_map
  Inst*  combinePredicates(Opnd* pred1, Opnd* pred2, Opnd* def,
                           Opnd* condition_pred);

  // Returns needed predicate for the current inst
  // If such predicate was already used, returns it
  // Otherwise create new pred along with andp inst
  Opnd*  calcInstPred(InstList& res_insts, Opnd* block_pred,
                      Opnd* inst_pred);
  Pred   toPred(const Opnd* opnd) const;
private:

  Ir*    _ir;
  Loop*  _loop;

  // { Pred1, Pred2 } -> PredDst means that PredDst = Pred1 && Pred2
  // Used to add predicate to already predicated insts
  PredMap _preds_map;
};

}      // namespace tango
#endif // TANGO_IF_CONVERSION_HPP
