#pragma once

#include "tango/compile_phase.hpp"
#include "tango/block_cloner.hpp"
#include "tango/inst_dependencies.hpp"
#include "tango/loop_transformations.hpp"
#include "tango/cfg.hpp"

namespace tango {
//----------------------------------------------------------------------------//

class LoopInvariants : public CompilePhase  {
public:
    LoopInvariants(Ir* ir) :
        CompilePhase(ir, PhaseName::MLI), cur_loop(NULL), dep_graph(NULL),
        clone_helper(ir), one_iteration_move(false), is_for_loop(false), has_invar_instruction(false),
        has_bound_check(false), do_store_localization(false), lan_ran(false), checkboundcond(NULL),
        ddg_builder(ir), rotate_base_moving(false), ltransform(ir) {}
    
    void                setLoop(Loop* _cur_loop) { cur_loop = _cur_loop; }
    void                setDepGraph(DepGraph* _dep_graph) { dep_graph = _dep_graph; }
    void                moveloopInvarDriver(Loop* );
    bool                checkValidLoop(Loop* );
    bool                hasInvariants(Loop* );
    void                doMoveLoopInvariants(Loop* );
    void                copyWithoutLoopInvariants(Loop*, Block*, bool);
    void                updateCondition(Inst*, int);
    MemLocalization*    getMemLoc(Inst *inst);
    Inst*               getinit(InstList initlist, Inst* inst);
    void                delete_memloc(Inst *inst);
    DepNode*            findEqualNode(DepNode* dn);
protected:
    virtual void run_internal();
    
private:
    Loop*               cur_loop;
    Block*              moveblock;
    DepGraph*           dep_graph;
    DepGraphBuilder     ddg_builder;
    CloneHelper         clone_helper;
    Inst*               final_val;
    Inst*               checkboundcond;
    Block*              checkblock;
    MemLocalizationList memlocs;
    DepNodeList         rotate_base_nodes;
    LoopTransformation  ltransform;
    bool                one_iteration_move;
    bool                is_for_loop;
    bool                has_bound_check;
    bool                do_store_localization;
    bool                has_invar_instruction;
    bool                rotate_base_moving;
    bool                lan_ran;
};

//----------------------------------------------------------------------------//

} // namespace tango
