#ifndef TANGO_LOOP_TRANSFORMATIONS
#define TANGO_LOOP_TRANSFORMATIONS

#include "tango/compile_phase.hpp"
#include "tango/block_cloner.hpp"
#include "tango/inst_dependencies.hpp"
#include "tango/pipeline_transformer.hpp"
#include "tango/cfg.hpp"

namespace tango {

//----------------------------------------------------------------------------//

class LoopTransformation : public CompilePhase {
public:
    LoopTransformation(Ir* ir) :
        CompilePhase(ir, PhaseName::LTR), cur_loop(NULL), dep_graph(NULL), clone_helper(ir), 
        pipeline_transformer(ir), ddg_builder(ir), unlinear_unroll_method(false) {}

    void                setLoop(Loop* _cur_loop) { cur_loop = _cur_loop; }
    void                setDepGraph(DepGraph* _dep_graph) { dep_graph = _dep_graph; }
    void                pipelineTransformDiver(Loop*);
    void                loopTransformDriver(Loop *);
    void                doUnroll(Block *bb, int unroll_cnt);
    void                doMultiBlockLoopUnroll(Loop *l, int unroll_cnt);
    void                doNonStandardLoopUnroll(Loop *l, int unroll_cnt);
    void                unrollUnstandardLoop(Loop *l, Block *new_head, Block *new_tail,int unroll_cnt);
    void                unrollLoopBlock(Block *b, Block *new_b, Block* post_block, Opnd* inroll_bound_reg, int unroll_cnt);
    void                unrollUnlinearLoop(Block *b, Block *new_b, Block* post_block, Opnd* inroll_bound_reg, int unroll_cnt);
    void                unrollMultiBlockLoop(Loop *l, Loop *new_l, Block *new_head, Block *new_tail,
                                              Block *post_b, Opnd* inroll_bound_reg, int unroll_cnt, Block* rem_b, Block *new_post);
    void                genPrePostBody(Block *head, Block *tail, Block *old_succ, Block *new_b,
                                       Block *new_l, Block *succ, RegOpnd* check_reg, int unroll_cnt, bool is_preloop);
    void                findPossibleDuplicateLoad(const InstList&, InstList::const_iterator, Inst*);
    bool                checkValidLoop(Loop* loop);
    bool                checkMultiBlock(Loop* loop);
    bool                checkInvariantOpnd(Opnd* opnd, Inst* inst, Loop* loop);
    bool                BlockIsAchievable(Block*, Block*, Loop* );
    int32_t             correctUnrollCntByPipelining(int32_t unroll_count);
    int                 calculateUnrollCount();
    bool                DefIsLiveAfterLoopExits(const RegOpnd* opnd, Loop* loop);
    bool                DefIsLiveAfterLoop(const RegOpnd* opnd, Loop* loop);
    bool                checkDefLiveAfterLoopExits(Inst* inst, Loop* loop);
    bool                checkDefLiveAfterLoop(Inst* inst, Loop* loop);
    bool                findReductions();
    bool                isPossibleDef(Block *b, Inst *inst);
    void                genOperandForInvariantReduction(Block* pre_block, int unroll_cnt);
    Block*              genRestoreBlock(Block*, Block*, int);
    void                do_unroll(Loop* loop);
protected:
    virtual void run_internal();
private:
    Loop*               cur_loop;
    DepGraph*           dep_graph;
    DepGraphBuilder     ddg_builder;
    CloneHelper         clone_helper;
    PipelineTransformer pipeline_transformer;
    OpndList         livedefs;
    OpndReplace         replaced;
    BlockReplace        blockreplaced;
    bool                unlinear_unroll_method;
};

//----------------------------------------------------------------------------//

}      // namespace tango
#endif // TANGO LOOP TRANSFROMATIONS
