#ifndef TANGO_ELBRUS_OPCODES_H
#define TANGO_ELBRUS_OPCODES_H

#define BIT(n) (1U << (n))

namespace tango {

//----------------------------------------------------------------------------//

enum ElbrusOpcode {

  // Некомбинированные операции с короткой кодировкой
  OPC_ADDd      = 0x11,    OPC_ADDs      = 0x10,    OPC_ANDd      = 0x01,    OPC_ANDNd     = 0x03,
  OPC_ANDNs     = 0x02,    OPC_ANDs      = 0x00,    OPC_BITREVd   = 0x27,    OPC_BITREVs   = 0x26,
  OPC_CCTOB     = 0x24,    OPC_CMPANDdb  = 0x23,    OPC_CMPANDsb  = 0x22,    OPC_CMPdb     = 0x21,
  OPC_CMPsb     = 0x20,    OPC_DRTOAP    = 0x62,    OPC_FADDd     = 0x31,    OPC_FADDs     = 0x30,
  OPC_FCMPdb    = 0x2f,    OPC_FCMPsb    = 0x2e,    OPC_FDTOD     = 0x3d,    OPC_FDTOS     = 0x3f,
  OPC_FMAXd     = 0x37,    OPC_FMAXs     = 0x36,    OPC_FMINd     = 0x35,    OPC_FMINs     = 0x34,
  OPC_FMULd     = 0x39,    OPC_FMULs     = 0x38,    OPC_FSTOD     = 0x3e,    OPC_FSTOS     = 0x3c,
  OPC_FSUBd     = 0x33,    OPC_FSUBs     = 0x32,    OPC_FXADDdd   = 0x41,    OPC_FXADDdx   = 0x43,
  OPC_FXADDss   = 0x40,    OPC_FXADDsx   = 0x42,    OPC_FXADDxd   = 0x45,    OPC_FXADDxs   = 0x44,
  OPC_FXADDxx   = 0x47,    OPC_FXCMPdb   = 0x29,    OPC_FXCMPsb   = 0x28,    OPC_FXCMPxb   = 0x2b,
  OPC_FXDIVdd   = 0x49,    OPC_FXDIVdx   = 0x4b,    OPC_FXDIVss   = 0x48,    OPC_FXDIVsx   = 0x4a,
  OPC_FXDIVTdd  = 0x61,    OPC_FXDIVTdx  = 0x63,    OPC_FXDIVTss  = 0x60,    OPC_FXDIVTsx  = 0x62,
  OPC_FXDIVxd   = 0x4d,    OPC_FXDIVxs   = 0x4c,    OPC_FXDIVxx   = 0x4f,    OPC_FXMULdd   = 0x51,
  OPC_FXMULdx   = 0x53,    OPC_FXMULss   = 0x50,    OPC_FXMULsx   = 0x52,    OPC_FXMULxd   = 0x55,
  OPC_FXMULxs   = 0x54,    OPC_FXMULxx   = 0x57,    OPC_FXRSUBdd  = 0x59,    OPC_FXRSUBdx  = 0x5b,
  OPC_FXRSUBss  = 0x58,    OPC_FXRSUBsx  = 0x5a,    OPC_FXSQRTIdx = 0x53,    OPC_FXSQRTIsx = 0x52,
  OPC_FXSQRTIxx = 0x57,    OPC_FXSQRTTdx = 0x5f,    OPC_FXSQRTTsx = 0x5e,    OPC_FXSQRTTxx = 0x5d,
  OPC_FXSQRTUdx = 0x5b,    OPC_FXSQRTUsx = 0x5a,    OPC_FXSQRTUxx = 0x59,    OPC_FXSUBdd   = 0x49,
  OPC_FXSUBdx   = 0x4b,    OPC_FXSUBss   = 0x48,    OPC_FXSUBsx   = 0x4a,    OPC_FXSUBxd   = 0x4d,
  OPC_FXSUBxs   = 0x4c,    OPC_FXSUBxx   = 0x4f,    OPC_GETFd     = 0x1f,    OPC_GETFs     = 0x1e,
  OPC_GETPL     = 0x63,    OPC_LDB       = 0x64,    OPC_LDCSB     = 0x68,    OPC_LDCSD     = 0x6b,
  OPC_LDCSH     = 0x69,    OPC_LDCSW     = 0x6a,    OPC_LDD       = 0x67,    OPC_LDDSB     = 0x6c,
  OPC_LDDSD     = 0x6f,    OPC_LDDSH     = 0x6d,    OPC_LDDSW     = 0x6e,    OPC_LDESB     = 0x70,
  OPC_LDESD     = 0x73,    OPC_LDESH     = 0x71,    OPC_LDESW     = 0x72,    OPC_LDFSB     = 0x74,
  OPC_LDFSD     = 0x77,    OPC_LDFSH     = 0x75,    OPC_LDFSW     = 0x76,    OPC_LDGSB     = 0x78,
  OPC_LDGSD     = 0x7b,    OPC_LDGSH     = 0x79,    OPC_LDGSW     = 0x7a,    OPC_LDH       = 0x65,
  OPC_LDSSB     = 0x7c,    OPC_LDSSD     = 0x7f,    OPC_LDSSH     = 0x7d,    OPC_LDSSW     = 0x7e,
  OPC_LDW       = 0x66,    OPC_LZCNTd    = 0x65,    OPC_LZCNTs    = 0x64,    OPC_MERGEd    = 0x0f,
  OPC_MERGEs    = 0x0e,    OPC_MOVFI     = 0x5c,    OPC_MOVIF     = 0x5e,    OPC_MOVTd     = 0x61,
  OPC_MOVTs     = 0x60,    OPC_MOVX      = 0x5f,    OPC_ORd       = 0x05,    OPC_ORNd      = 0x07,
  OPC_ORNs      = 0x06,    OPC_ORs       = 0x04,    OPC_POPCNTd   = 0x67,    OPC_POPCNTs   = 0x66,
  OPC_SARd      = 0x1d,    OPC_SARs      = 0x1c,    OPC_SCLd      = 0x15,    OPC_SCLs      = 0x14,
  OPC_SCRd      = 0x17,    OPC_SCRs      = 0x16,    OPC_SDIVd     = 0x43,    OPC_SDIVs     = 0x42,
  OPC_SDIVX     = 0x46,    OPC_SHLd      = 0x19,    OPC_SHLs      = 0x18,    OPC_SHRd      = 0x1b,
  OPC_SHRs      = 0x1a,    OPC_SMODX     = 0x47,    OPC_STB       = 0x24,    OPC_STCSB     = 0x28,
  OPC_STCSD     = 0x2b,    OPC_STCSH     = 0x29,    OPC_STCSW     = 0x2a,    OPC_STD       = 0x27,
  OPC_STDSB     = 0x2c,    OPC_STDSD     = 0x2f,    OPC_STDSH     = 0x2d,    OPC_STDSW     = 0x2e,
  OPC_STESB     = 0x30,    OPC_STESD     = 0x33,    OPC_STESH     = 0x31,    OPC_STESW     = 0x32,
  OPC_STFSB     = 0x34,    OPC_STFSD     = 0x37,    OPC_STFSH     = 0x35,    OPC_STFSW     = 0x36,
  OPC_STGSB     = 0x38,    OPC_STGSD     = 0x3b,    OPC_STGSH     = 0x39,    OPC_STGSW     = 0x3a,
  OPC_STH       = 0x25,    OPC_STSSB     = 0x3c,    OPC_STSSD     = 0x3f,    OPC_STSSH     = 0x3d,
  OPC_STSSW     = 0x3e,    OPC_STW       = 0x26,    OPC_SUBd      = 0x13,    OPC_SUBs      = 0x12,
  OPC_SXT       = 0x0c,    OPC_UDIVd     = 0x41,    OPC_UDIVs     = 0x40,    OPC_UDIVX     = 0x44,
  OPC_UMODX     = 0x45,    OPC_VFSI      = 0x63,    OPC_XORd      = 0x09,    OPC_XORNd     = 0x0b,
  OPC_XORNs     = 0x0a,    OPC_XORs      = 0x08,

  // Расширения кода операции
  OPX_EXT       = 0x01,
  OPX_FLB       = 0x04,    OPX_FLH       = 0x05,    OPX_FLW       = 0x06,    OPX_FLD       = 0x07,
  OPX_ICMB0     = 0x08,    OPX_ICMB1     = 0x09,    OPX_ICMB2     = 0x0a,    OPX_ICMB3     = 0x0b,
  OPX_FCMB0     = 0x0c,    OPX_FCMB1     = 0x0d,    OPX_PFCMB0    = 0x0e,    OPX_PFCMB1    = 0x0f,
  OPX_FXCMB     = 0x14,

  // Некомбинированные операции с длинной кодировкой
  OPL_APTOAP    = 0x50,    OPL_APTOAPB   = 0x51,    OPL_CAST      = 0x55,    OPL_FCMPd     = 0x2d,
  OPL_FCMPdf    = 0x2f,    OPL_FCMPs     = 0x2c,    OPL_FCMPsf    = 0x2e,    OPL_FDIVd     = 0x49,
  OPL_FDIVs     = 0x48,    OPL_FRCPs     = 0x50,    OPL_FRSQRTs   = 0x54,    OPL_FSQRTId   = 0x4d,
  OPL_FSQRTs    = 0x4c,    OPL_FSQRTTd   = 0x51,    OPL_FXCMPdf   = 0x29,    OPL_FXCMPsf   = 0x28,
  OPL_FXCMPxf   = 0x2b,    OPL_GETSOD    = 0x5a,    OPL_GETSP     = 0x58,    OPL_GETTAGd   = 0x09,
  OPL_GETTAGs   = 0x08,    OPL_GETTC     = 0x24,    OPL_GETTD     = 0x56,    OPL_GETVA     = 0x52,
  OPL_INVTC     = 0x26,    OPL_LDAAB     = 0x5c,    OPL_LDAAD     = 0x5f,    OPL_LDAAH     = 0x5d,
  OPL_LDAAQ     = 0x7f,    OPL_LDAAW     = 0x5e,    OPL_LDAPB     = 0x68,    OPL_LDAPD     = 0x6b,
  OPL_LDAPH     = 0x69,    OPL_LDAPQ     = 0x7a,    OPL_LDAPW     = 0x6a,    OPL_LDCSQ     = 0x42,
  OPL_LDCUDB    = 0x60,    OPL_LDCUDD    = 0x63,    OPL_LDCUDH    = 0x61,    OPL_LDCUDQ    = 0x78,
  OPL_LDCUDW    = 0x62,    OPL_LDDSQ     = 0x43,    OPL_LDESQ     = 0x44,    OPL_LDFSQ     = 0x45,
  OPL_LDGDB     = 0x64,    OPL_LDGDD     = 0x67,    OPL_LDGDH     = 0x65,    OPL_LDGDQ     = 0x79,
  OPL_LDGDW     = 0x66,    OPL_LDGSQ     = 0x46,    OPL_LDODPB    = 0x74,    OPL_LDODPD    = 0x77,
  OPL_LDODPH    = 0x75,    OPL_LDODPQ    = 0x7d,    OPL_LDODPW    = 0x76,    OPL_LDODRB    = 0x6c,
  OPL_LDODRD    = 0x6f,    OPL_LDODRH    = 0x6d,    OPL_LDODRQ    = 0x7b,    OPL_LDODRW    = 0x6e,
  OPL_LDODWB    = 0x70,    OPL_LDODWD    = 0x73,    OPL_LDODWH    = 0x71,    OPL_LDODWQ    = 0x7c,
  OPL_LDODWW    = 0x72,    OPL_LDRD      = 0x5b,    OPL_LDSSQ     = 0x47,    OPL_MOVTq     = 0x57,
  OPL_MULd      = 0x21,    OPL_MULs      = 0x20,    OPL_ODTOAP    = 0x54,    OPL_PACKSSHB  = 0x08,
  OPL_PACKSSWH  = 0x0a,    OPL_PACKUSHB  = 0x09,    OPL_PADDB     = 0x08,    OPL_PADDD     = 0x0f,
  OPL_PADDH     = 0x09,    OPL_PADDSB    = 0x0a,    OPL_PADDSH    = 0x0b,    OPL_PADDUSB   = 0x0c,
  OPL_PADDUSH   = 0x0d,    OPL_PADDW     = 0x0e,    OPL_PANDd     = 0x48,    OPL_PANDNd    = 0x49,
  OPL_PAVGUSB   = 0x1e,    OPL_PAVGUSH   = 0x1f,    OPL_PCMPEQB   = 0x18,    OPL_PCMPEQH   = 0x19,
  OPL_PCMPEQW   = 0x1a,    OPL_PCMPGTB   = 0x1b,    OPL_PCMPGTH   = 0x1c,    OPL_PCMPGTW   = 0x1d,
  OPL_PEXTRH    = 0x1e,    OPL_PFADDd    = 0x31,    OPL_PFADDs    = 0x30,    OPL_PFCMPd    = 0x3b,
  OPL_PFCMPs    = 0x3a,    OPL_PFDIVd    = 0x4b,    OPL_PFDIVs    = 0x4a,    OPL_PFMAXd    = 0x37,
  OPL_PFMAXs    = 0x36,    OPL_PFMINd    = 0x35,    OPL_PFMINs    = 0x34,    OPL_PFMULd    = 0x39,
  OPL_PFMULs    = 0x38,    OPL_PFSQRTs   = 0x4e,    OPL_PFSQRTTd  = 0x53,    OPL_PFSUBd    = 0x33,
  OPL_PFSUBs    = 0x32,    OPL_PINSH     = 0x1f,    OPL_PMADDH    = 0x1a,    OPL_PMAXSH    = 0x03,
  OPL_PMAXUB    = 0x02,    OPL_PMINSH    = 0x01,    OPL_PMINUB    = 0x00,    OPL_PMOVMSKB  = 0x0b,
  OPL_PMOVMSKPD = 0x07,    OPL_PMOVMSKPS = 0x06,    OPL_PMULHH    = 0x18,    OPL_PMULHUH   = 0x1b,
  OPL_PMULLH    = 0x19,    OPL_PMULUBHH  = 0x1d,    OPL_PORd      = 0x4a,    OPL_PSADBW    = 0x1c,
  OPL_PSHUFH    = 0x17,    OPL_PSHUFW    = 0x16,    OPL_PSLLD     = 0x4e,    OPL_PSLLH     = 0x15,
  OPL_PSLLQH    = 0x0e,    OPL_PSLLQL    = 0x0f,    OPL_PSLLW     = 0x14,    OPL_PSRAH     = 0x13,
  OPL_PSRAW     = 0x12,    OPL_PSRLD     = 0x4c,    OPL_PSRLH     = 0x11,    OPL_PSRLQH    = 0x0c,
  OPL_PSRLQL    = 0x0d,    OPL_PSRLW     = 0x10,    OPL_PSUBB     = 0x10,    OPL_PSUBD     = 0x17,
  OPL_PSUBH     = 0x11,    OPL_PSUBSB    = 0x12,    OPL_PSUBSH    = 0x13,    OPL_PSUBUSB   = 0x14,
  OPL_PSUBUSH   = 0x15,    OPL_PSUBW     = 0x16,    OPL_PUNPCKHBH = 0x00,    OPL_PUNPCKHHW = 0x02,
  OPL_PUNPCKHWD = 0x04,    OPL_PUNPCKLBH = 0x01,    OPL_PUNPCKLHW = 0x03,    OPL_PUNPCKLWD = 0x05,
  OPL_PUTTAGd   = 0x0b,    OPL_PUTTAGs   = 0x0a,    OPL_PUTTC     = 0x25,    OPL_PXORd     = 0x4b,
  OPL_RRd       = 0x3f,    OPL_RRs       = 0x3e,    OPL_RWd       = 0x3d,    OPL_RWs       = 0x3c,
  OPL_SMULH     = 0x71,    OPL_SMULX     = 0x23,    OPL_UMULH     = 0x70,    OPL_UMULX     = 0x22,
  OPL_STAAB     = 0x1c,    OPL_STAAD     = 0x1f,    OPL_STAAH     = 0x1d,
  OPL_STAAQ     = 0x3f,    OPL_STAAW     = 0x1e,    OPL_STAPB     = 0x28,    OPL_STAPD     = 0x2b,
  OPL_STAPH     = 0x29,    OPL_STAPQ     = 0x3a,    OPL_STAPW     = 0x2a,    OPL_STCSQ     = 0x02,
  OPL_STDSQ     = 0x03,    OPL_STESQ     = 0x04,    OPL_STFSQ     = 0x05,    OPL_STGDB     = 0x24,
  OPL_STGDD     = 0x27,    OPL_STGDH     = 0x25,    OPL_STGDQ     = 0x39,    OPL_STGDW     = 0x26,
  OPL_STGSQ     = 0x06,    OPL_STODPB    = 0x34,    OPL_STODPD    = 0x37,    OPL_STODPH    = 0x35,
  OPL_STODPQ    = 0x3d,    OPL_STODPW    = 0x36,    OPL_STODRB    = 0x2c,    OPL_STODRD    = 0x2f,
  OPL_STODRH    = 0x2d,    OPL_STODRQ    = 0x3b,    OPL_STODRW    = 0x2e,    OPL_STODWB    = 0x30,
  OPL_STODWD    = 0x33,    OPL_STODWH    = 0x31,    OPL_STODWQ    = 0x3c,    OPL_STODWW    = 0x32,
  OPL_STRD      = 0x1b,    OPL_STSSQ     = 0x07,    OPL_TDTOMP    = 0x53,

  // Комбинированные операции ICMB0, ICMB1, ICMB2 (1 стадия)
  OPM_1_AND  = 0x00,       OPM_1_ANDN    = 0x01,    OPM_1_OR      = 0x02,    OPM_1_ORN     = 0x03,
  OPM_1_XOR  = 0x04,       OPM_1_XORN    = 0x05,    OPM_1_RSRVD   = 0x06,    OPM_1_MERGE   = 0x07,
  OPM_1_ADD  = 0x08,       OPM_1_SUB     = 0x09,    OPM_1_SCL     = 0x0a,    OPM_1_SCR     = 0x0b,
  OPM_1_SHL  = 0x0c,       OPM_1_SHR     = 0x0d,    OPM_1_SAR     = 0x0e,    OPM_1_GETF    = 0x0f,

  // Комбинированные операции ICMB0 (2 стадия)
  OPM_20_AND = 0x00,       OPM_20_ANDN   = 0x01,    OPM_20_OR     = 0x02,    OPM_20_ORN    = 0x03,

  // Комбинированные операции ICMB1 (2 стадия)
  OPM_21_XOR = 0x00,       OPM_21_XORN   = 0x01,    OPM_21_RSUB   = 0x02,    OPM_21_RSRVD  = 0x03,

  // Комбинированные операции ICMB2 (2 стадия)
  OPM_22_ADD = 0x00,       OPM_22_SUB    = 0x01,    OPM_22_RSRVD1 = 0x02,    OPM_22_RSRVD2 = 0x03,

  // Комбинированные операции ICMB3 (Операция INSF)
  OPM_31_INSF = 0x06,      OPM_3_INSFd   = 0x6d,

  // Комбинированные операции FCMB0 (1nd)
  OPF_1_FADD  = 0x00,      OPF_1_FSUB    = 0x01,    OPF_1_RSRVD1  = 0x02,    OPF_1_RSRVD2  = 0x03,
  OPF_1_FMUL  = 0x04,      OPF_1_RSRVD3  = 0x05,    OPF_1_RSRVD4  = 0x06,    OPF_1_RSRVD5  = 0x07,

  // Комбинированные операции FCMB0 (2nd)
  OPF_2_FADD  = 0x00,      OPF_2_FSUB    = 0x01,    OPF_2_RSRVD1  = 0x02,    OPF_2_RSRVD2  = 0x03,

  // Комбинированные операции FCMB1 (2nd)
  OPF_12_RSRVD1  = 0x00,      OPF_12_FRSUB    = 0x01,    OPF_12_RSRVD2  = 0x02,    OPF_12_RSRVD3  = 0x03,

  // Комбинированные операции PFCMB0 (2nd)

  // Комбинированные операции PFCMB1 (1nd)
  OPP1_PFADD = 0x0,       OPP1_PFSUB    = 0x01,    OPP1_PFHADD = 0x03,    OPP1_PFHSUB   = 0x03,
  OPP1_PFMUL = 0x4,       OPP1_RSRVD    = 0x05,    OPP1_PSHUFB = 0x06,    OPP1_PFADDSUB = 0x07,
                                                   OPP1_PMERGE = 0x06,
  // Комбинированные операции PFCMB1 (2nd)
  OPP2_RSRVD = 0x0,       OPP2_PFRSUB   = 0x01,    OPP2_PSHUFB = 0x02,    OPP2_PFADDSUB = 0x03,
                                                                          OPP2_PMERGE   = 0x03,

  OPP_PSHUFB = 0x4d,      OPP_PMERGE    = 0x6d,

  // Операции над целыми с результатом в виде флага
  OPF_ANDs      = 0x00,    OPF_ANDNs     = 0x02,    OPF_ORs       = 0x04,    OPF_ORNs      = 0x06,
  OPF_XORs      = 0x08,    OPF_XORNs     = 0x0a,    OPF_ADDs      = 0x10,    OPF_SUBs      = 0x12,
  OPF_SCLs      = 0x14,    OPF_SCRs      = 0x16,    OPF_SHLs      = 0x18,    OPF_SHRs      = 0x1a,
  OPF_SARs      = 0x1c,    OPF_UMULX     = 0x22,    OPF_SMULX     = 0x23,    OPF_INCs      = 0x30,
  OPF_DECs      = 0x32,

  // Операции слога CS0
  OC0_IBRANCH   = 0x0,     OC0_DONE      = 0x3,     OC0_DISP      = 0x0,     OC0_LDISP     = 0x1,
  OC0_SDISP     = 0x2,     OC0_RETURN    = 0x3,

  // Операции слога CS1
  OC1_SETR0     = 0x00,    OC1_SETR1     = 0x01,    OC1_SETEI     = 0x02,    OC1_WAIT      = 0x03,
  OC1_SETBR     = 0x04,    OC1_CALL      = 0x05,    OC1_MAS       = 0x06,    OC1_FLUSHR    = 0x07,

  // Операции целых сравнений с результатом в виде предиката
  OCI_OVW       = 0x00,    OCI_LTU       = 0x01,    OCI_EQ        = 0x02,    OCI_LEU       = 0x03,
  OCI_NEG       = 0x04,    OCI_ODD       = 0x05,    OCI_LTS       = 0x06,    OCI_LES       = 0x07,

  // Операции вещественных сравнений с результатом в виде предиката
  OCF_EQ        = 0x00,    OCF_LT        = 0x01,    OCF_LE        = 0x02,    OCF_UO        = 0x03,
  OCF_NE        = 0x04,    OCF_NLT       = 0x05,    OCF_NLE       = 0x06,    OCF_OD        = 0x07,

  // Операции сравнения вещественных с результатом в виде битовой маски
  OCFM_EQ       = 0xc0,    OCFM_LT       = 0xc1,    OCFM_LE       = 0xc2,    OCFM_UO       = 0xc3,
  OCFM_NE       = 0xc4,    OCFM_NL       = 0xc5,    OCFM_NLE      = 0xc6,    OCFM_OD       = 0xc7,

  // Операции сравнения вещественных с результатом в виде флагов
  OCFF_OD       = 0xc7,    OCFF_UO       = 0xc3,

  // Операции преобразования вещественных/целых
  OCV_FDTOFS    = 0xc6,    OCV_FDTOFX    = 0xc7,    OCV_FDTOID    = 0xc0,    OCV_FDTOIDTR  = 0xc2,
  OCV_FDTOIS    = 0xc0,    OCV_FDTOISTR  = 0xc2,    OCV_FSTOFD    = 0xc6,    OCV_FSTOFX    = 0xc7,
  OCV_FSTOID    = 0xc0,    OCV_FSTOIDTR  = 0xc2,    OCV_FSTOIS    = 0xc0,    OCV_FSTOISTR  = 0xc2,
  OCV_FXTOFD    = 0xc6,    OCV_FXTOFS    = 0xc7,    OCV_FXTOID    = 0xc1,    OCV_FXTOIDTR  = 0xc3,
  OCV_FXTOIS    = 0xc1,    OCV_FXTOISTR  = 0xc3,    OCV_IDTOFD    = 0xc4,    OCV_IDTOFS    = 0xc4,
  OCV_IDTOFX    = 0xc5,    OCV_ISTOFD    = 0xc4,    OCV_ISTOFS    = 0xc4,    OCV_ISTOFX    = 0xc5,
  OCV_PFDTOFS   = 0xce,    OCV_PFDTOIS   = 0xc8,    OCV_PFDTOISTR = 0xca,    OCV_PFSTOFD   = 0xce,
  OCV_PFSTOIS   = 0xc8,    OCV_PFSTOISTR = 0xca,    OCV_PISTOFS   = 0xcc,    OCV_FDTOFXTOFD= 0xcf,
  OCV_FSTOFXTOFS= 0xcf,

  // Операции преобразования адресов
  OPA_NONE      = 0xc0,    OPA_CUD       = 0xf0,    OPA_GD        = 0xf2,    OPA_USD       = 0xec,
  OPA_STAT      = 0xee,

  // Операции Move для буфера предподкачки массива (AAS.opc)
  OAA_MOVAB     = 0x01,    OAA_MOVAH     = 0x02,    OAA_MOVAW     = 0x03,    OAA_MOVAD     = 0x04,
  OAA_MOVAQ     = 0x05,

  // Коды операций поля ALS.opce операций MOVT
  OPM_MV        = 0xc0,    OPM_MVC       = 0xc1,    OPM_MVR       = 0xc2,    OPM_MVRC      = 0xc3,

  // Коды операций поля ALS.opce операций INVTC
  OIV_TCA       = 0xc0,    OIV_TCE       = 0xc1,

  // Коды операций для значений по умолчанию полей ALS.opce и ALES.opce
  OPC_NONE      = 0xc0,

  // Коды операций поля ALS.opce операций MOVX
  OMX_MV        = 0xc0,    OMX_MVXA      = 0xc1,    OMX_MVXC      = 0xc2,

  // Коды операций SXT (SRC1)
  OSXT_BS       = 0x00,    OSXT_HS       = 0x01,    OSXT_WS       = 0x02,
  OSXT_BZ       = 0x04,    OSXT_HZ       = 0x05,    OSXT_WZ       = 0x06,

  // Коды операций полуслога CD
  OCD_RLP       = 0x00,    OCD_MRGC      = 0x02,

  // Коды операций слога PLS
  OPL_CLPAND    = 0x00,    OPL_CLPLAND   = 0x01,    OPL_MLP       = 0x03,

  OPC_LAST
};

enum ElbrusCURegister {
  PSR = 0x0,
  WD = 0x1,
  CWD = 0x6,
  PSP_hi = 0x7,
  PSP_lo = 0x9,
  PSHTP = 0x0b,
  PCSP_hi = 0x0d,
  PCSP_lo = 0x0f,
  PCSHTP = 0x13,
  CTPR1 = 0x15,
  CTPR2 = 0x16,
  CTPR3 = 0x17,
  USBR = 0x1e,
  CUTD = 0x21,
  EIR = 0x23,
  TSD = 0x24,
  CUIR = 0x25,
  OSCUD_hi = 0x26,
  OSCUD_lo = 0x27,
  OSGD_hi = 0x28,
  OSGD_lo = 0x29,
  OSEM = 0x2a,
  USD_hi = 0x2c,
  USD_lo = 0x2d,
  TR = 0x2e,
  OSR0 = 0x2f,
  CUD_hi = 0x30,
  CUD_lo = 0x31,
  GD_hi = 0x32,
  GD_lo = 0x33,
  CS_hi = 0x34,
  CS_lo = 0x35,
  DS_hi = 0x36,
  DS_lo = 0x37,
  ES_hi = 0x38,
  ES_lo = 0x39,
  FS_hi = 0x3a,
  FS_lo = 0x3b,
  GS_hi = 0x3c,
  GS_lo = 0x3d,
  SS_hi = 0x3e,
  SS_lo = 0x3f,
  DIBCR = 0x40,
  DIMCR = 0x41,
  DIBSR = 0x42,
  DTCR = 0x43,
  DIBAR0 = 0x48,
  DIBAR1 = 0x49,
  DIBAR2 = 0x4a,
  DIBAR3 = 0x4b,
  DIMAR0 = 0x4c,
  DIMAR1 = 0x4d,
  DTARF = 0x4e,
  DTART = 0x4f,
  CF_CWD_0__hi = 0x51,
  CF_CWD_0__lo = 0x53,
  CF_CWD_16__hi = 0x55,
  CF_CWD_16__lo = 0x57,
  CU_HW0 = 0x78,
  UPSR = 0x80,
  IP = 0x81,
  nIP = 0x82,
  LSR = 0x83,
  PFPFR = 0x84,
  FPCR = 0x85,
  FPSR = 0x86,
  ILCR = 0x87,
  BR = 0x88,
  BGR = 0x89,
  IDR = 0x8a,
  CLKR = 0x90,
  TIR_hi = 0x9c,
  TIR_lo = 0x9d,
  RPR_lo = 0xa0,
  SBBP = 0xa1,
  RPR_hi = 0xa2
};

enum ALUNum {
  ALU_0 = 0,
  ALU_1,
  ALU_2,
  ALU_3,
  ALU_4,
  ALU_5,

  ALU_TOTAL
};

#define USD_BASE_MASK 0x0000FFFFFFFFFFFFULL

//----------------------------------------------------------------------------//

} // namespace tango
#endif // TANGO_ELBRUS_OPCODES_H
