#ifndef TANGO_PASS_MANAGER_HPP
#define TANGO_PASS_MANAGER_HPP

#include <mutex>
#include <iomanip>

#include "tango/tango_inst.hpp"
#include "tango/atomic/atomic.hpp"

//-------------------------------------------------------------------------------------//

namespace tango {

//-------------------------------------------------------------------------------------//
// Synchronize output, generate id for session's
class PassManager : public PersistentObj {
 public:
  static PassManager& getInstance() { static PassManager manager; return manager; }

  void          setLogOutputType()     { _output.resetOutputType(TangoLogSinglethread); }
  size_t        generateId()           { atomic::inc((int32_t*)&_s_id); return _s_id;   }
  bool          isSingleThread()       { return _output._single_thread; }
  TangoStream&  getOutput()            { return _output;                }
  std::mutex&   getMutex()             { return _mutex;                 }
  void          flush(arena_stringstream& oss) {
    _mutex.lock();
    _output << oss.str();
    _mutex.unlock();
  }
  void          trackPhaseTime(PhaseName phase_name, double time) {
    int phase_num = static_cast<int>(phase_name);
    _mutex.lock();
    _phase_comp_times[phase_num] += time;
    _mutex.unlock();
  }

  ~PassManager() {
    if (TangoCITime) {
      const int MAX_LEN = 40;
      // TODO: reuse these names from global logger
      char phases[][MAX_LEN] = {
        "",
        "Block_Cloner",
        "Code_Layouter",
        "Copy_Coalesscer",
        "Copy_Propagation",
        "CondReadRegs",
        "Dead_Code_Elimination",
        "Dynamic_Checks_Generator",
        "Emitter",
        "Fast_Copy_Coalesscer",
        "If_Conversion",
        "Instruction_Scheduler",
        "Live_Analyser",
        "Loop_Building",
        "LandPeep",
        "Loop_Transformations",
        "Move_Invariants_Outside_Loop",
	      "Blocks_Merge",
        "OOP_Resolver",
        "Peephole",
        "Prolog_Generator",
        "Register_Allocator",
        "SSA_Dead_Code_Elemination",
        "Selector",
        "State_Flusher",
        "Spill/Fill_Generator",
        "Software_Pipeliner",
        "DeSSA",
        "SXTCorrector",
        "Tree_Balancer",
        "Cold_Path_Merging",
        "Macro_Instructions"
      };
      for (uint i = 1; i < PhaseName::END - 1; i++) {
        std::cerr << std::setw(37) << std::left << phases[i]
                  << _phase_comp_times[i] << "s\n";
      }
    }
  }

 private:
  // required for singletone
  PassManager() : _s_id(0) {
    for (uint i = 0; i < PhaseName::END - 1; i++) {
      _phase_comp_times[i] = 0.0;
    }
  }

 private:
  TangoStream             _output;
  double                  _phase_comp_times[PhaseName::END - 1];

  uint32_t                _s_id; //session id
  std::mutex              _mutex;
};

//-------------------------------------------------------------------------------------//

}      // namespace tango
#endif // TANGO PASS MANAGER
