#pragma once

#include "tango/inst_dependencies.hpp"
#include "tango/if_conversion.hpp"
#include "tango/cfg.hpp"

namespace tango {

//----------------------------------------------------------------------------//

const float PTR_MINIMAL_PROFILE = 5.0f;

// This class transforms loop to a single block
// Loop must contain no calls inside and be a superblock
// with exits
class PipelineTransformer : public TangoObj {
public:
  PipelineTransformer(Ir* ir) : _ir(ir), _if_conversion(ir) {}

  // Transforms loop to a single block if possible
  // Returns true if transformation finished successfully
  bool   transform(Loop* loop);

private:
  bool   checkPreconditions(Loop* loop) const;
  void   buildLoopInfo(Loop* loop); // Init internal fields
  bool   canBeTransformed() const;
  bool   hasFinalRedefinition() const;
  bool   checkAntiDefs() const;
  bool   needToBeTransformed() const;
  bool   isSuperLoop() const;       // Check that _loop is a superblock with exits

  void   transformInternal();
  void   updateExitPreds() const;
  Block* getNext(const Block* block) const; // Get next block in superloop's trace

  void   splitLastBlock() const;
  // Redirect early exits to new tail and generate exit determining after loop
  void   moveEarlyExits();
  // Adds pred = false in the loop beginning
  void   nullifyLoopPred() const;
  void   removeRedundantPredicates() const;

private:
  Ir*      _ir;
  Loop*    _loop;

  Block*   _main_exit;   // Loop tail's exit
  BlockSet _exit_blocks; // Set of early exit blocks from loop
                         // NB: one exit block can have several preds in loop

  InstList _exit_transfers; // List of cts determining the right exit after the
                            // loop

  LoopIfConversion _if_conversion;
};

//----------------------------------------------------------------------------//

} // namespace tango
