#ifndef RDF_H
#define RDF_H

#include "tango/ir.hpp"
#include "tango/post_dominators.hpp"

namespace tango {

typedef arena_map<int32_t, BlockList>::type  int322BlockListMap;

// RDF for block i is a set of blocks such that for each block j from this set
// one path from i leads to block j and one doesn't.
// Full description of algorithm can be found in
// Keith D. Cooper, Timothy J. Harvey, and Ken Kennedy.
// A Simple, Fast Dominance Algorithm.
//
// Example where blocks j and k are included in RDF(block i) (e1, e2 - exit nodes):
//            i
//          /   \
//         j     k
//        / \     \
//       e1  e2   ...
//
class ReverseDominanceFrontier : public TangoObj {
public:
  void               build(PostDomTree& dt);
  BlockList&         frontiers(int32_t block_id);
  
private:
  int322BlockListMap _frontiers;
};

} // namespace tango

#endif // RDF_H
