#ifndef TANGO_REGISTER_ALLOCATOR
#define TANGO_REGISTER_ALLOCATOR

#include "tango/compile_phase.hpp"

namespace tango {

//----------------------------------------------------------------------------//

class RegisterAllocator : public CompilePhase {
public:
  RegisterAllocator(Ir* ir) : CompilePhase(ir, PhaseName::RAR),
    predicate_reg_alloc_flag(ir->_PredicateRegAllocInScheduler) {}

protected:
  void         buildMatrix();
  void         defLiveSet(Inst*, float);
  void         useLiveSet(Inst*, float);
  void         useTouchLiveSet(Inst*, float);
  void         updateDepSet(RegOpnd*);
  
  void         assignLocations();
  void         assignLocations(RegOpndVector&);
  void         assignLocation(RegOpnd*);
  void         buildFreeMask(RegOpnd*);
  void         correctSpillCosts();

  void         printMatrix();

protected:
  virtual void run_internal();

protected:
  LiveSet      live_set;      // live set used for data flow iteration
  BitMap       cross_call_site_regs;
  bool         predicate_reg_alloc_flag;
  bool         before_stack_pointer;
};

//----------------------------------------------------------------------------//

}      // namespace tango
#endif // TANGO REGISTER ALLOCATOR
