#ifndef TANGO_SPILL_GENERATOR_HPP
#define TANGO_SPILL_GENERATOR_HPP

#include "tango/compile_phase.hpp"

namespace tango {

//----------------------------------------------------------------------------//

typedef arena_vector<int32_t>::type IntVector;
typedef arena_map<int32_t, Inst*>::type IntInstMap;

//----------------------------------------------------------------------------//

class SpillGen : public CompilePhase {
public:
  SpillGen(Ir* ir) : CompilePhase(ir, PhaseName::SGN), spill_regs(), max_spill(0) {}

  static const int SPILL_REG_NUM = 3;

  void        processBundle(BundleList&, BundleList::iterator&);
  void        stealRegs(uint need_regs);
  void        generateFills(Bundle*);
  void        generateSpills(Bundle*);
  void        generateRestoreStealed(IntVector& to_restore, BundleList& bundles, BundleList::iterator& it);
  void        needSpillOrFill(Bundle*, int&, int&, bool&, bool&);
  Opnd*       spill(RegOpnd*, bool);
  int32_t     getStackOffset(int32_t location);
  void        spillRegR(RegOpnd*, Opnd*);
  void        fillRegR(RegOpnd*, Opnd*);
  void        initSpillRegs();
  int32_t     getAvailableSpillReg(RegType);
  Opnd*       alreadyFilledReg(int32_t);

protected:
  virtual void run_internal();

public:
  IntVector   spill_regs;
  IntVector   stealed_regs;
  IntVector   stealed_stack_slots;
  InstList    spill_code;
  InstList    fill_code;
  IntInstMap  restore_stealed_code;
  InstList    steal_code;
  int32_t     max_spill;

  BitMap      uniq_fill_set;
  BitMap      regmask;
};

//----------------------------------------------------------------------------//

}      // namespace tango
#endif // TANGO SPILL GENERATOR HPP
