#ifndef SSA_DCE
#define SSA_DCE

#include "tango/compile_phase.hpp"
#include "tango/rdf.hpp"
#include "tango/post_dominators.hpp"

namespace tango {

//
// Dead code eliminator on SSA-form. Uses algorithm from
// Ron Cytron, Jeanne Ferrante, Barry K. Rosen, Mark N. Wegman, F. Kenneth Zadeck
// Efficiently computing static single assignment form and the control dependence graph
// It also marks defs for insts from touch uses as live (only for safepoints).
//
class SsaDeadCodeEliminator: public CompilePhase {
public:
  SsaDeadCodeEliminator(Ir* ir): CompilePhase(ir, PhaseName::SDCE) {}

protected:
  virtual void run_internal();

private:

  void   mark(ReverseDominanceFrontier &rdf) const;
  void   sweep(PostDomTree &postdom);

  // mark methods
  void   addCritical(InstList &worklist) const;
  void   addOpndsDefs(InstList &worklist, const OpndVector &opnds) const;
  void   addRdfs(InstList &worklist, const BlockList &rdfs) const;
  bool   isCritical(const Inst *inst) const;
  bool   isDefined(const RegOpnd *opnd) const;

  // sweep methods
  Block* findFirstAlive(Block *current, PostDomTree &postdom) const;
  bool   isAlive(const Block *block) const;
  Inst*  branchToAliveBlock(Block *block, PostDomTree &postdom) const;
};

} // namespace tango

#endif // SSA_DCE
