#ifndef TANGO_GLOBALS_HPP
#define TANGO_GLOBALS_HPP

#ifndef assert_rvm
#include <cassert>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>

#ifndef PRODUCT
# define TANGO_ASSERT 1

# ifdef ASSERT
#  ifndef TANGO_ASSERT
#   define TANGO_ASSERT 1
#  endif
# endif

# ifndef assert_rvm
#  define assert_rvm(p, msg) assert((p) && (msg))
# endif
#else
# ifndef assert_rvm
#  define assert_rvm(p, msg)
# endif
#endif

// This check added to avoid annoying warnings in release build about DEBUG==1 check
#undef __DEBUG_IS_EQUAL_TO_ONE
#ifdef DEBUG
# if DEBUG == 1
#  define __DEBUG_IS_EQUAL_TO_ONE
# endif // DEBUG==1
#endif // DEBUG

#if defined(SLOW_PRODUCT) || !defined(PRODUCT) || defined(__DEBUG_IS_EQUAL_TO_ONE)
#define TANGO_LOG_ENABLED
#define TANGO_LOG(code) code
#else
#define TANGO_LOG(code)
#endif

#if !defined(PRODUCT) || defined(__DEBUG_IS_EQUAL_TO_ONE)
#define TANGO_VERIVIER_ENABLED
#endif

#undef __DEBUG_IS_EQUAL_TO_ONE

// PRODUCT variant
#ifdef PRODUCT
#define PRODUCT_ONLY(code) code
#define NOT_PRODUCT(code)
#define NOT_PRODUCT_ARG(arg)
#define PRODUCT_RETURN  {}
#define PRODUCT_RETURN0 { return 0; }
#define PRODUCT_RETURN_(code) { code }
#else // PRODUCT
#define PRODUCT_ONLY(code)
#define NOT_PRODUCT(code) code
#define NOT_PRODUCT_ARG(arg) arg,
#define PRODUCT_RETURN  /*next token must be ;*/
#define PRODUCT_RETURN0 /*next token must be ;*/
#define PRODUCT_RETURN_(code)  /*next token must be ;*/
#endif // PRODUCT

#ifdef TANGO_ASSERT
#define DEBUG_ONLY(code) code
#define NOT_DEBUG(code)
#define NOT_DEBUG_RETURN  /*next token must be ;*/
// Historical.
#define debug_only(code) code
#else // TANGO_ASSERT
#define DEBUG_ONLY(code)
#define NOT_DEBUG(code) code
#define NOT_DEBUG_RETURN {}
#define debug_only(code)
#endif // TANGO_ASSERT

#define NULL_WORD 0L

typedef unsigned char* address;

#define TANGO_LIB_FLAGS(flag)                                                                                  \
  flag(TangoLog,                     const char*, "",    "Print Tango CG logs")                                 \
  flag(TangoLogDiag,                 bool,        false, "Print map Options from GlobalLogger")                 \
  flag(AllowCopyForSuperblocks,      bool,        true,  "Allow to create clones to make superblocks longer")   \
  flag(SuperblockScheduling,         bool,        true,  "Chose trace of hot blocks and process them as one")   \
  flag(ForbidCreateCompCode,         bool,        true,  "Forbid to move code down if it requires comp code")   \
  flag(UsePredsForStores,            bool,        true,  "Use predicate registers to move stores throw clf")    \
  flag(Verbose,                      bool,        true,  "")                                                    \
  flag(TangoUseCopyPropagation,      bool,        true,  "Use copy propagation optimization")                   \
  flag(TangoLoopsBuild,              bool,        true,  "Generate loops in Tango")                            \
  flag(TangoIfConversion,            bool,        true,  "If Conversion in Tango")                              \
  flag(TangoInstructionScheduler,    bool,        true,  "Use tango scheduler")                                 \
  flag(GenerateDynamicChecks,        bool,        true,  "Creates dynamic checks memory intersections in loops") \
  flag(PredicateRegAllocInScheduler, bool,        true,  "")                                                    \
  flag(TangoUnroll,                  bool,        true,  "One block loops are unrolled in Tango")               \
  flag(OptReduction,                 bool,        true,  "Optimization for loop reductions")                    \
  flag(UseFloatReduction,            bool,        false,  "allow Optimization for float reductions")            \
  flag(OptMemoryLocalization,        bool,        true,  "Localization of memory instructions insid loop")     \
  flag(UnrollAfterPipeline,          bool,        true,  "Unroll performs after Pipelining")                   \
  flag(OptRotateBaseMoving,          bool,        true, "Transformation of array base address for memory instructions") \
  flag(RAW_SXTProcessing,            bool,        true, "processing RAW_SXT instruction for linear form evaluation") \
  flag(TangoMoveInvariants,          bool,        true,  "Move Invariants outside Loop in Tango")               \
  flag(TangoCheckBoundOpt,           bool,        true,  "Check Bounds Optimizaion in Tango")                   \
  flag(TangoMultiBlocksLoop,         bool,        true,  "Tango procecess multi blocks loops")                  \
  flag(UseMemoryDisambiguation,      bool,        true,  "Memory disambiguation in scheduler")                  \
  flag(UseSoftwarePipelining,        bool,        true,  "Pipeline loops")                                      \
  flag(UseLockCheckDisambiguation,   bool,        false, "Divide loads to locks and checks for disambiguation") \
  flag(UsePipelineTransformer,       bool,        false, "Transform superloops to single block")                \
  flag(TangoLogSinglethread,         bool,        true,  "Type of how logger print logs")                       \
  flag(TangoGreedyScheduler,         bool,        false, "Use greedy scheduler")                                \
  flag(UseLandp,                     bool,        false,  "Use LANDP instruction to fold branches")              \
  flag(TangoSsaDce,                  bool,        false, "Use SSA based DCE algorythm")                         \
  flag(TangoCITime,                  bool,        false, "collect timing information for compilation")          \
  flag(DisamLdStLimit,               int32_t,     20, "Maximum amount of stores which can load cross")          \
  flag(TangoCmpBlockLimit,           int32_t,     40, "Maximum amount of CMP in one superblock")                \
  flag(UseBrRegs,                    bool,        true, "Use BR register for allocation")                       \
  flag(UseVerifier,                  bool,        false, "Tango Verifier")                                      \
  flag(TreeHeightBalancing,          bool,        false, "Tree height balancing")                               \
  flag(TangoColdPathMerging,         bool,        false, "Cold Path Merging")                                   \
  flag(TangoCmpIBranchMacro,         bool,        false, "Use CMP-IBRANCH Macro Instructions")                  \
  flag(TangoInsertIBranches,         bool,        false, "Use IBRANCH Instructions for cold paths")                  \
  flag(TangoCombinedArith,           bool,        false, "Use Combined E2K Instructions")                           \
  flag(PredicateAnalyses,            bool,        false, "Handle instructions under predicate in live analyser phase") \
  flag(DisableMova,                  bool,        false, "Disable MOVA instructions for hw bug workaround on v6")

namespace tango {
#define TANGO_FLAG_DECL(FLAG_NAME, TYPE, DEFAULT_VALUE, DESCRIPTION) extern TYPE FLAG_NAME;
TANGO_LIB_FLAGS(TANGO_FLAG_DECL)
}

namespace tango {
const int LogBytesPerShort   = 1;
const int LogBytesPerInt     = 2;
const int LogBytesPerWord    = 3;
const int LogBytesPerLong    = 3;

const int BytesPerShort      = 1 << LogBytesPerShort;
const int BytesPerInt        = 1 << LogBytesPerInt;
const int BytesPerWord       = 1 << LogBytesPerWord;
const int BytesPerLong       = 1 << LogBytesPerLong;

const int LogBitsPerByte     = 3;
const int LogBitsPerShort    = LogBitsPerByte + LogBytesPerShort;
const int LogBitsPerInt      = LogBitsPerByte + LogBytesPerInt;
const int LogBitsPerWord     = LogBitsPerByte + LogBytesPerWord;
const int LogBitsPerLong     = LogBitsPerByte + LogBytesPerLong;

const int BitsPerByte        = 1 << LogBitsPerByte;
const int BitsPerShort       = 1 << LogBitsPerShort;
const int BitsPerInt         = 1 << LogBitsPerInt;
const int BitsPerWord        = 1 << LogBitsPerWord;
const int BitsPerLong        = 1 << LogBitsPerLong;

const intptr_t AllBits    = ~0; // all bits set in a word
const intptr_t NoBits     =  0; // no bits set in a word
const int64_t  NoLongBits =  0; // no bits set in a long
const intptr_t OneBit     =  1; // only right_most bit set in a word

#define nth_bit(n)        (((n) >= BitsPerWord) ? 0 : (OneBit << (n)))
#define right_n_bits(n)   (nth_bit(n) - 1)
#define left_n_bits(n)    (right_n_bits(n) << (((n) >= BitsPerWord) ? 0 : (BitsPerWord - (n))))

const int wordSize = sizeof(intptr_t);
const uint32_t max_uint32 = (uint32_t)-1;

#define align_size_up_(size, alignment) (((size) + ((alignment) - 1)) & ~((alignment) - 1))

class HeapWord {
 private:
  char* i;
};

inline bool is_size_aligned(size_t size, size_t alignment) {
  return align_size_up_(size, alignment) == size;
}

inline bool is_ptr_aligned(void* ptr, size_t alignment) {
  return align_size_up_((intptr_t)ptr, (intptr_t)alignment) == (intptr_t)ptr;
}

inline size_t align_size_up(size_t size, intptr_t alignment) {
  return align_size_up_(size, alignment);
}

#define align_size_down_(size, alignment) ((size) & ~((alignment) - 1))

inline intptr_t align_size_down(intptr_t size, intptr_t alignment) {
  return align_size_down_(size, alignment);
}

void print_and_crash(const char* file, int line, const char* msg);

#ifndef ShouldNotReachHere
# define ShouldNotReachHere() tango::print_and_crash(__FILE__, __LINE__, "ShouldNotReachHere");
#endif

#ifndef ShouldNotCallThis
# define ShouldNotCallThis() tango::print_and_crash(__FILE__, __LINE__, "ShouldNotCallThis");
#endif

// TODO: this method belongs in assembler
inline int wsz_to_wbs(int wsz) {
  return ((wsz * 2) - 8) / 2;
}

} // namespace tango

#endif

