#ifndef SHARE_VM_TANGO_VERIFIER_HPP
#define SHARE_VM_TANGO_VERIFIER_HPP

#include "tango/global_logger.hpp"

namespace tango {

typedef arena_map<int32_t, RegOpndVector>::type LocationFlowMap;
typedef arena_map<Opnd*, const Block*>::type    SsaDefMap;
typedef arena_vector<LabelOpnd*>::type          LabelVector;
typedef arena_set<int32_t>::type                OpndIdSet;
typedef arena_set<void*>::type                  RelocAddrSet;
typedef arena_set<Reloc*>::type                 RelocSet;

//-------------------------------------------------------------------------------------//

class Verifier : public TangoObj {
public:
  Verifier(Ir*);

  void    run();

  void    verifyCfg();
  void    verifySSA();
  void    verifyDataFlow();
  void    verifyInsts();
  void    verifyRelocs();
  void    verifyCodeLayout();
  void    verifyTerminators();

  void    checkLocationFlow();

  void    reportAndCrush();

protected:
  //relocations verification
  void    checkReloc               (Inst*, RelocSet&, RelocAddrSet&, OpndSet&);

  // CFG verification
  void    checkBlock               (const Block*);
  void    checkLabels              (const Block*);
  void    checkBlockList           (const Block*);
  void    checkPreds               (const Block*);
  void    checkSuccs               (const Block*);
  void    checkBrTargets           (const Block*);
  void    checkBrTarget            (const Block*, Opnd*);
  void    checkEdges               (const Block*);
  bool    findBr                   (const Block*, const Block*);

  //SSA verification
  void    checkSsaDefUse           (SsaDefMap&);
  void    checkSsaUse              (const Inst*, const SsaDefMap&, bool);
  void    checkPhi                 (const Block*, const Inst*, SsaDefMap&);
  void    checkPhiUse              (RegOpnd*, LabelOpnd*, const Block*, SsaDefMap&);
  void    checkPhiUseFlags         (const Block*, const Inst*, OpndVector&);
  bool    isInBranch               (const Block*, const Block*);

  //Dataflow check
  void    checkOpnd                (Inst*, RegOpnd*);
  void    checkCtpr                (Inst*, RegOpnd*);
  void    checkDefUse              (const Block*, const Inst*, OpndSet&);
  void    checkCtprDefUse          (const Block*);

  //Inst check
  void    checkInst                (Inst*);

  void    checkCall                (Inst*);
  void    checkSetwd               (Inst*);
  void    checkGetsp               (Inst*);

  void    checkDisp                (Inst*);
  void    checkLdisp               (Inst*);
  void    checkSdisp               (Inst*);
  void    checkMovtd               (Inst*);
  void    checkReturn              (Inst*);

  void    checkBranch              (Inst*);
  void    checkIbranch             (Inst*);
  void    checkUnlockSt            (Inst*);
  void    checkCtCond              (Inst*);
  void    checkCtRet               (Inst*);
  void    checkSwitch              (Inst*);

  // check location flow
  void    checkLocationFlowInsts   (Block*);
  void    checkLocationFlowBundles (Block*);
  void    defLocationFlow          (Block*, Inst*);
  void    useLocationFlow          (Block*, Inst*);
  void    locationError            (Block*, Inst*, RegOpnd*);
  void    mergeOpndVectors         (Block*);

  //terminators check before layouter
  void    checkTerminators         (Block*);
  void    check2Terminators        (Block*, InstVector&);

  bool    isTerminator             (Inst*);
  bool    isJumpOrCall             (InstCode);

public:
  Ir*                             _ir;
  arena_stringstream              _report;

private:
  RegOpndVector                   _cur_opnd_set;
  LocationFlowMap                 _location_map;

  bool                            _after_ssa;
  bool                            _after_layouter;
  bool                            _after_regalloc;

  bool                            _has_setwd;
  bool                            _reg_use;
};

//-------------------------------------------------------------------------------------//



//-------------------------------------------------------------------------------------//

}      // namespace tango
#endif // SHARE VM TANGO VERIFIER
